"""
**String manipulation and translation helper module**

-----------------
"""
import pathlib
import re


def str2intb(bytes, default=None):
    if isinstance(bytes, int):
        return bytes

    bytes = bytes and bytes.strip()

    if not bytes:
        return default

    try:
        int_bytes = bytes if len(bytes) > 0 and bytes[0] != b"0" or bytes.startswith(b"0x") else b"0x%s" % bytes
        return int(int_bytes, 0)
    except ValueError:
        return bytes if default is None else default


def str2int(str, default=None):
    if isinstance(str, int):
        return str

    str = str and str.strip()

    if not str:
        return default

    try:
        int_str = str if len(str) > 0 and str[0] != "0" or str.startswith("0x") else "0x%s" % str
        return int(int_str, 0)
    except ValueError:
        return str if default is None else default


def cast(_val, encoding="utf-8", errors="surrogateescape"):
    if isinstance(_val, bytes):
        return typebytes(encoding, errors)
    if isinstance(_val, str):
        return typestr(encoding, errors)
    if isinstance(_val, bytearray):
        return typebytearray(encoding, errors)
    if isinstance(_val, memoryview):
        return typememoryview(encoding, errors)
    if isinstance(_val, pathlib.PurePath):
        return typestr(encoding, errors)
    if hasattr(_val, "path"):
        return cast(_val.path, encoding, errors)
    raise TypeError("invalid parameter type")


def typebytes(encoding="utf-8", errors="surrogateescape"):
    class wrapper(bytes):
        def __new__(cls, val: str | bytes):
            if isinstance(val, bytes):
                return val
            try:
                return val.encode(encoding, errors=errors)
            except AttributeError:
                if isinstance(val, (bytearray, memoryview)):
                    return bytes(val)
                return val

        revert = typestr(encoding=encoding, errors=errors)

    return wrapper


def typestr(encoding="utf-8", errors="surrogateescape"):
    class wrapper(str):
        def __new__(cls, val: str | bytes):
            if isinstance(val, str):
                return val
            try:
                return val.decode(encoding, errors=errors)
            except AttributeError:
                if isinstance(val, memoryview):
                    return val.tobytes().decode(encoding, errors=errors)
                return val

    return wrapper


def typebytearray(encoding="utf-8", errors="surrogateescape"):
    class wrapper(bytearray):
        def __new__(cls, val: str | bytes):
            val = typebytes(encoding, errors)(val)
            try:
                return bytearray(val)
            except (AttributeError, UnicodeDecodeError):
                return val

    return wrapper


def typememoryview(encoding="utf-8", errors="surrogateescape"):
    def wrapper(val: str | bytes):
        val = typebytes(encoding, errors)(val)
        try:
            return memoryview(val)
        except (AttributeError, UnicodeDecodeError):
            return val

    return wrapper


def hr_size_toint(str):
    if str is None or str == "" or isinstance(str, (int, float)):
        return str

    gr = re.match(r"([\d\.]+)(?:([BKMGTbkmgt])(i)?[bB]?)?", str)
    if not gr:
        raise ValueError('can not parse string "%s" in bytes' % str)

    val, unit, bi = gr
    val = str2int(val)
    unit = unit.lower() if unit else "b"
    bi = 1024 if bi else 1000

    return val * {"b": 1, "k": bi, "m": bi**2, "g": bi**3, "t": bi**4}.get(unit)


def hr_size(size: int) -> str:
    """
    convert byte size to human readable string
    """
    for unit in ["", "Ki", "Gi"]:
        if abs(size) < 1024:
            break
        size = int(size / 1024)

    return "%.1f%sB" % (size, unit if size < 1024 else "T")


# copy from client.py
def bytestohostport(host, port=None):
    # host contains port only
    try:
        return b"", int(host)
    except ValueError:
        pass

    i = host.rfind(b":")
    j = host.rfind(b"]")  # ipv6 addresses have [...]
    if i > j:
        try:
            port = int(host[i + 1 :])
        except ValueError:
            if host[i + 1 :] != "":  # http://foo.com:/ == http://foo.com/
                raise ValueError("nonnumeric port: '%s'" % host[i + 1 :]) from None
        host = host[:i]
    else:
        port = port

    if host and host[0] == "[" and host[-1] == "]":
        host = host[1:-1]

    return (host, port)



from ..interfaces import (
    ABCDelegation,
    ICoreFileSystemService,
    ICoreStreamService,
    ICoreShellService,
)

__all__ = ["ShellService", "FileSystemService", "StreamService"]


class FileSystemService(ICoreFileSystemService, ABCDelegation):
    __delegated__ = (ICoreFileSystemService,)

    def __init__(self, ifs):
        self.__delegate__(ICoreFileSystemService, ifs)

    def list(self, path: bytes, *, extras: dict | None = None) -> list[bytes]:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.list(path, extras=extras)

    def stat(self, path: bytes, *, extras: dict | None = None) -> dict:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.stat(path)

    def rm(self, path: bytes, *, extras: dict | None = None) -> None:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.rm(path, extras=extras)

    def mkdir(self, path: bytes, *, extras: dict | None = None) -> None:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.mkdir(path, extras=extras)

    def rmdir(self, path: bytes, *, extras: dict | None = None) -> None:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.rmdir(path, extras=extras)

    def is_container(self, path: bytes, *, extras: dict | None = None) -> bool:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.is_container(path, extras=extras)

    def exist(self, path: bytes, *, extras: dict | None = None) -> bool:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.exist(path, extras=extras)


def _assert_env(env):
    if env is not None:
        if not hasattr(env, "items"):
            return (
                False,
                "invalid env type detected, env must be of type dict not %s" % (env.__class__.__name__),
            )
        for k, v in env.items():
            if not isinstance(k, bytes):
                return (
                    False,
                    "invalid key type detected in env, argument must be of type bytes not %s" % (k.__class__.__name__),
                )
            if not isinstance(v, bytes):
                return (
                    False,
                    "invalid value type detected in env, argument must be of type bytes not %s" % (v.__class__.__name__),
                )
    return True, "ok"


def _assert_cmd_bytes(cmd):
    return (
        isinstance(cmd, bytes),
        f"command argument must be of type bytes not {cmd.__class__.__name__}",
    )


def _assert_working_dir(working_dir):
    return (
        working_dir is None or isinstance(working_dir, bytes),
        f"working_dir argument must be of type bytes not {working_dir.__class__.__name__}",
    )


def _assert_cmd_list(cmd):
    if len(cmd) == 0:
        return False, "invalid command value, command must not be empty"

    if not isinstance(cmd, (list, tuple)):
        return (
            False,
            f"command argument must be of type bytes not {cmd.__class__.__name__}",
        )

    for arg in cmd:
        if not isinstance(arg, bytes):
            return (
                False,
                f"invalid argument type detected in command, argument must be of type bytes not {arg.__class__.__name__}",
            )

    return True, "ok"


class ShellService(ICoreShellService, ABCDelegation):
    __delegated__ = (ICoreShellService,)

    def __init__(self, ishell):
        self.__delegate__(ICoreShellService, ishell)

    def execute(
        self,
        command: bytes,
        working_dir: bytes | None = None,
        env: dict | None = None,
        *,
        extras: dict | None = None,
    ) -> tuple[bytes, bytes, bytes]:
        assert all(m := _assert_cmd_bytes(command)), m
        assert all(m := _assert_working_dir(working_dir)), m
        assert all(m := _assert_env(env)), m

        return self._delegate_.execute(command, working_dir, env, extras=extras)

    def spawn(
        self,
        command: bytes,
        working_dir: bytes | None = None,
        env: dict | None = None,
        *,
        extras: dict | None = None,
    ) -> int:
        assert all(m := _assert_cmd_list(command)), m
        assert all(m := _assert_working_dir(working_dir)), m
        assert all(m := _assert_env(env)), m

        return self._delegate_.spawn(command, working_dir, env)


class StreamService(ICoreStreamService, ABCDelegation):
    __delegated__ = (ICoreStreamService,)

    def __init__(self, istream):
        self.__delegate__(ICoreStreamService, istream)

    def open_file(self, path: bytes, wiring: int, *, extras: dict | None = None) -> tuple[bytes, int]:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.open_file(path, wiring, extras=extras)

    def open_process(
        self,
        path: bytes,
        wiring: int,
        args: list[bytes],
        working_dir: bytes | None = None,
        env: dict[bytes, bytes] | None = None,
        *,
        extras: dict | None = None,
    ) -> tuple[bytes, int, int, int, int]:
        assert isinstance(path, bytes), f"path argument must be of type bytes not {path.__class__.__name__}"

        return self._delegate_.open_process(path, wiring, args, working_dir, env, extras=extras)

    def write(self, handle: int, data: bytes, *, extras: dict | None = None):
        assert isinstance(data, (bytes, bytearray)), f"data argument must be of type bytes not {data.__class__.__name__}"

        return self._delegate_.write(handle, data, extras=extras)



import importlib
import typing

from myrrh.core.services import load_ext_group
from ..myrrhc import cmd, types  # noqa: F401
from ..myrrhc.cmd import myrrhc_cmds  # noqa: F401


_extensions: dict[str, typing.Any] = dict()


def getsession():
    from myrrh.tools.myrrhc.session import getsession

    return getsession()


def load_commands(name):
    _extensions[name] = importlib.import_module(name)


load_ext_group("myrrh.tools.myrrhc_ext")



from ...interfaces import IMHandle, IRuntimeObject

__all__ = ("MHandle",)


class MHandle(IMHandle):
    __slots__ = ("val", "_omgr", "_closed", "obj")

    def __new__(cls, val, omgr):
        cls = type(
            cls.__name__,
            (cls,),
            {"obj": omgr.geto(val), "val": int(val), "_omgr": omgr, "_closed": False},
        )

        return object.__new__(cls)

    def __enter__(self):
        return self

    def __exit__(self, _t, _v, _tb):
        self.close()

    def __int__(self):
        return self.val

    def __del__(self):
        try:
            if not self._closed:
                self.close()
        # flake8: noqa: E722
        except:
            pass

    def __repr__(self):
        return f'{self.__class__.__name__}({int(self)}{":closed" if self.closed else ""})->{str(self.obj)}'

    def __getattr__(self, attr):
        return getattr(self.obj, attr)

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            return super().__setattr__(attr, value)

        return setattr(self.obj, attr, value)

    def close(self):
        if not self._closed:
            self._omgr.close(self)
            self._closed = True

    def detach(self) -> int:
        self._closed = True
        return int(self)

    @property
    def closed(self):
        return self._closed or self.obj.closed


IRuntimeObject.register(MHandle)  # type: ignore[type-abstract]



from myrrh.provider.provider_ext import provider, cmd


@provider.group()
def local():
    ...


@local.command()
@cmd.option("--cwd", default=None, type=cmd.Path())
@cmd.argument("name")
def new(cwd, name):
    import bmy

    bmy.new("**/local", eid=name, cwd=cwd)



import shutil
import traceback
import warnings
import sys
import os
import re
import logging

import pkg_resources
import logging.config
import json

from importlib import metadata

__all__ = [
    "rebase",
    "cfg_get",
    "cfg_set",
    "cfg_del",
    "__distname__",
    "__version__",
    "__copyright__",
    "__license__",
]

myrrh_sys = "myrrh.msys"
__distname__ = "myrrh"

try:
    from ...__license__ import copyright as __copyright__, license as __license__
except Exception:
    __copyright__ = ""
    __license__ = ""

_JSON_INDENT = 2

PID = os.getpid()


def _load(f):
    return json.load(f)


def _save(f, cfg):
    json.dump(cfg, f, indent=_JSON_INDENT)


def myrrh_default_cfg_dirs():
    try:
        from click import utils

        cfg_dir = utils.get_app_dir(__distname__)
    except ImportError:
        cfg_dir = os.path.expanduser(os.path.join("~", "." + __distname__))

    if not os.path.exists(cfg_dir):
        try:
            os.makedirs(cfg_dir)
        except Exception:
            pass  # ignore errors => no config dir

    return [cfg_dir]


def myrrh_versioning(file, dist_name):
    try:
        # try to find version from package dist
        version = pkg_resources.get_distribution(dist_name).version
    except Exception:
        version = "0.0.0"

    if version == "0.0.0":
        # try to find version from scm
        try:
            from setuptools_scm import get_version

            return get_version(relative_to=file)
        except Exception:
            pass

    return version


try:
    from ...__version__ import version as __version__
except Exception:
    __version__ = myrrh_versioning(__file__, __distname__)


def _msysfile(fpath):
    msys = {}
    with open(fpath) as f:
        with open(fpath) as f:
            msys = _load(f)

    if not isinstance(msys, dict):
        raise IOError(f"{fpath} incompatible configuration file type, dictionary required")

    msys["@mbase@"] = os.path.abspath(fpath)

    getattr(sys, "__msys__").update(msys)

    return getattr(sys, "__msys__")


def _msys(paths=None):
    if not paths:
        paths = myrrh_default_cfg_dirs()

    for path in paths:
        fpath = os.path.join(path, myrrh_sys) if (os.path.isdir(path)) else path
        if os.path.isfile(fpath):
            try:
                _msysfile(fpath)
            except Exception as e:
                warnings.warn(f'file "{fpath}" found but is invalid: {str(e)}, try to find another one')
            else:
                break
    else:
        try:
            fpath = os.path.join(paths[0], myrrh_sys)
            if not os.path.exists(fpath):
                with open(fpath, "w") as f:
                    _save(f, dict())

                _msysfile(fpath)
        except Exception:
            fpath = None

    if fpath:
        log = logging.getLogger("myrrh")
        log.info(f"system variables found at {fpath}")

    return getattr(sys, "__msys__")


def _rebase_internal(path=None):
    __msys__ = getattr(sys, "__msys__", {})

    path = path or __msys__.get("@mbase@")
    installed_extensions = __msys__.get("@installed_extensions@") or list()
    entities = __msys__.get("@entities@") or list()

    setattr(
        sys,
        "__msys__",
        {
            "@mversion@": __version__,
            "@mpath@": os.path.dirname(__file__),
            "@copyright@": __copyright__,
            "@license@": __license__,
            "@etc@": myrrh_default_cfg_dirs(),
            "@metadata@": dict(metadata.metadata("myrrh")),
            "@loaded_extensions@": [],
            "@failed_extensions@": [],
            "@installed_extensions@": installed_extensions,
            "@actived_ext_groups@": [],
            "@entities@": entities,
        },
    )

    cwd_file = os.path.join(os.getcwd(), myrrh_sys)

    if path and os.path.isfile(path):
        result = _msysfile(path)

    elif path and os.path.isdir(path):
        result = _msys([path])
    elif os.path.isfile(cwd_file):
        result = _msysfile(cwd_file)
    else:
        result = _msys()

    return result


_CONF_OBJ_TYPE = (
    "loggers",
    "handlers",
    "formatters",
    "filters",
)


def configure_log():
    import configparser

    logging_cfg = cfg_get(section="logging")

    cfg = configparser.RawConfigParser()

    for obj_type in _CONF_OBJ_TYPE:
        cfg.add_section(obj_type)
        cfg[obj_type]["keys"] = logging_cfg.get(obj_type) or ""
        keys = filter(None, re.split(" *, *| +| *; *", cfg[obj_type]["keys"]))
        for key in keys:
            obj_name = obj_type.rstrip("s")
            cfg.add_section(f"{obj_name}_{key}")
            for k, v in cfg_get(section=f"logging.{obj_name}.{key}").items():
                cfg[f"{obj_name}_{key}"][k] = v

    logging.config.fileConfig(cfg, disable_existing_loggers=logging_cfg.get("disable_existing_loggers", True))

    if not cfg_get("enable", True, section="logging"):
        logging.disable()


def rebase(path=None):
    global log

    extension_bkup = sys.__msys__["@loaded_extensions@"] + sys.__msys__["@failed_extensions@"] if hasattr(sys, "__msys__") else None
    extension_groups = sys.__msys__["@actived_ext_groups@"] if hasattr(sys, "__msys__") else None

    result = _rebase_internal(path=None)

    if cfg_get("rebase", False):
        result = _rebase_internal(cfg_get("rebase"))

    cfg_init("rebase", "")

    cfg_init("raise_on_failure", False, section="extensions")
    cfg_init("search_for", True, section="extensions")
    cfg_init("default_value", True, section="extensions")

    cfg_init("enable", False, section="logging")
    cfg_init("version", 1, section="logging")
    cfg_init("incremental", False, section="logging")
    cfg_init("disable_existing_loggers", True, section="logging")
    cfg_init("loggers", "root, myrrh", section="logging")
    cfg_init("handlers", "console, myrrh", section="logging")
    cfg_init("formatters", "myrrh", section="logging")
    cfg_init("filters", "", section="logging")

    cfg_init("level", "CRITICAL", section="logging.logger.root")
    cfg_init("handlers", "console", section="logging.logger.root")

    cfg_init("level", "INFO", section="logging.logger.myrrh")
    cfg_init("handlers", "myrrh", section="logging.logger.myrrh")
    cfg_init("qualname", "myrrh", section="logging.logger.myrrh")
    cfg_init("propagate", "0", section="logging.logger.myrrh")

    cfg_init("class", "StreamHandler", section="logging.handler.myrrh")
    cfg_init("level", "DEBUG", section="logging.handler.myrrh")
    cfg_init("formatter", "myrrh", section="logging.handler.myrrh")
    cfg_init("args", "(sys.stdout,)", section="logging.handler.myrrh")

    cfg_init("class", "StreamHandler", section="logging.handler.console")
    cfg_init("level", "CRITICAL", section="logging.handler.myrrh")
    cfg_init("args", "(sys.stdout,)", section="logging.handler.myrrh")

    cfg_init(
        "format",
        "%(relativeCreated).1f - %(name)s:%(threadName)s - %(levelname)s - %(message)s",
        section="logging.formatter.myrrh",
    )

    log.info("myrrh new session")

    find_ext()

    if extension_groups:
        for group in extension_groups:
            load_ext_group(group)

    if extension_bkup:
        for extension in extension_bkup:
            load_ext(extension)

    return {k: v for k, v in result.items() if not (k.startswith("@") and k.endswith("@"))}


def _cfg_get_section(d, section):
    if not section:
        return d

    if not section.startswith(("__", "@__")):
        section = "__" + section
    if not section.endswith(("__", "__@")):
        section = section + "__"

    if not d.get(section):
        d[section] = dict()

    return d[section]


def _persistent_setcfg(cfg_path, cfg):
    try:
        backup = f"{cfg_path}.bak"
        if os.path.isfile(cfg_path):
            shutil.copy(cfg_path, backup)
        else:
            with open(backup, "w") as f:
                _save(f, cfg)
    except Exception as e:
        log.warning(f"Failed to backup msys config in {cfg_path}: {str(e)}")

    try:
        with open(cfg_path, "w") as f:
            _save(f, cfg)
    except Exception as e:
        log.info(f"Failed to save msys config in {cfg_path}: {str(e)}, config will not be persistante")


def _persistent_getcfg(cfg_path):
    try:
        with open(cfg_path) as f:
            fcfg = _load(f)

        return fcfg
    except Exception as e:
        log.info(f"Failed to load msys config in{cfg_path}: {str(e)}, try restore backup")

    try:
        with open(f"{cfg_path}.bak") as f:
            fcfg = _load(f)

        _persistent_setcfg(cfg_path, fcfg)

        return fcfg

    except Exception as e:
        log.info(f"Failed to load backup msys config in{cfg_path}: {str(e)}, config is lost...")

    return


def cfg_init(key, value, section=""):
    return cfg_set(key, value, section, overwrite=False)


def cfg_set(key, value, section="", *, overwrite=True):
    cfg_path = sys.__msys__.get("@mbase@", "")
    fcfg = None

    if section.startswith("@"):
        raise ValueError('section name could not start with "@" character')

    if key.startswith(("@", "__")):
        raise ValueError('key name could not start with "__" or "@" character')

    if cfg_path:
        fcfg = _persistent_getcfg(cfg_path)

    cfg = sys.__msys__ if fcfg is None else fcfg

    if overwrite or key not in _cfg_get_section(cfg, section):
        _cfg_get_section(cfg, section)[key] = value

        if fcfg is not None:
            _persistent_setcfg(cfg_path, cfg)

        if fcfg and cfg_path:
            _rebase_internal(cfg_path)

    return _cfg_get_section(cfg, section)[key]


def cfg_del(key, section=""):
    cfg_path = sys.__msys__.get("@mbase@", "")
    fcfg = None

    if section.startswith("@"):
        raise ValueError(f'section "{key}" is defined as a runtime section and could not be deleted')

    if key.startswith("@"):
        raise ValueError(f'key "{key}" is defined as a runtime key and could not be deleted')

    if cfg_path:
        fcfg = _persistent_getcfg(cfg_path)

    cfg = sys.__msys__ if fcfg is None else fcfg
    _cfg_get_section(cfg, section).pop(key, None)

    if fcfg is not None:
        _persistent_setcfg(cfg_path, cfg)

    if fcfg is not None and cfg_path:
        _rebase_internal(cfg_path)


def cfg_get(key="", default=None, *, section=""):
    if not key:
        return dict(_cfg_get_section(sys.__msys__, section))

    return _cfg_get_section(sys.__msys__, section).get(key, default)


# decorator
def cfg_prop(section=None, key=None):
    def wrapper(func):
        func.section = _section_name(func) if section is None else section
        func.key = key or func.__name__
        func.default = cfg_init(func.key, func(), section=func.section)

        def _get(self):
            return cfg_get(func.key, default=func.default, section=func.section)

        def _set(self, value):
            cfg_set(func.key, value, section=func.section)

        return property(_get, _set)

    return wrapper


def _section_name(obj):
    import inspect

    mod = inspect.getmodule(obj)
    if not mod:
        return ""

    return mod.__package__ or mod.__name__


class DynCfg:
    def __new__(cls, *a, **kwa):
        if not hasattr(cls, "__SECTION__"):
            cls.__SECTION__ = _section_name(cls)

        return super().__new__(cls, *a, **kwa)

    def all(self):
        return cfg_get(section=self.__SECTION__)

    def init(self, key, value):
        return cfg_init(key, value, section=self.__SECTION__)

    def get(self, key, default=""):
        return cfg_get(key, default=default, section=self.__SECTION__)

    def set(self, key, value=""):
        return cfg_set(key, value, section=self.__SECTION__)


def find_ext():
    if not cfg_get("search_for", section="__extensions__", default=True):
        return

    myrrh_exts = {name: exts for name, exts in metadata.entry_points().items() if name.startswith(f"{__distname__}.")}
    sys.__msys__["@installed_extensions@"] = []
    for _, exts in myrrh_exts.items():
        for ext in exts:
            sys.__msys__["@installed_extensions@"].append(_extension_string(ext.group, ext.name, ext.value))


def _extension_string(group, name, value):
    return f"{group}/{name}={value}"


def _extension_partition(extension):
    group, _, extension = extension.partition("/")
    name, _, value = extension.partition("=")

    return group, name, value


def load_ext(extension, name_value=None):
    cfg = cfg_get(section="__extensions__")

    if name_value:
        group = extension
        name, value = name_value
    else:
        group, name, value = _extension_partition(extension)

    extension = _extension_string(group, name, value)

    if extension in sys.__msys__["@loaded_extensions@"]:
        return

    if extension in sys.__msys__["@failed_extensions@"]:
        sys.__msys__["@failed_extensions@"].remove(extension)

    log.debug(f"load extension: {group}, {name}")

    try:
        callable = metadata.EntryPoint(name, f"{group}:{value}", group).load()
        callable(*name.split("-"))
        sys.__msys__["@loaded_extensions@"].append(f"{extension}")
    except (ModuleNotFoundError, ImportError, Exception) as e:
        sys.__msys__["@failed_extensions@"].append(f"{extension}")
        if cfg.get("raise_on_failure"):
            raise
        log.debug(f"load extension failure : {group}, {name}")
        warnings.warn(f"{''.join(traceback.format_tb(e.__traceback__))}\nFailed to load {name}, this extension will be unavailable: {str(e)}")


def load_ext_group(group):
    sys.__msys__["@actived_ext_groups@"].append(group)

    if not (cfg_get(group, section="__extensions__") or cfg_get("default_value", True, section="__extensions__")):
        return

    ext_cfg = cfg_get(section=f"ext.{group}")

    for name, value in ext_cfg.items():
        load_ext(group, (name, value))

    for installed in sys.__msys__["@installed_extensions@"]:
        if installed.startswith(group):
            load_ext(installed)


#
log = logging.getLogger("myrrh")
rebase()



import gettext
import os

from errno import *  # noqa: F403
import warnings

default_errorno = 15

posix_error_mappings = {
    "No such file or directory": ENOENT,  # noqa : F405
    "not found": ENOENT,  # noqa : F405
    "Permission denied": EACCES,  # noqa : F405
    "Too many levels of symbolic links": ELOOP,  # noqa : F405
    "Not enough space": ENOMEM,  # noqa : F405
    "Not a directory": ENOTDIR,  # noqa : F405
    "Operation not permitted": EPERM,  # noqa : F405
    "No such process": ESRCH,  # noqa : F405
    "Try again": EAGAIN,  # noqa : F405
    "Too many open files": EMFILE,  # noqa : F405
    "File too large": EFBIG,  # noqa : F405
    "Broken pipe": EPIPE,  # noqa : F405
    "Directory nonexistent": ENOENT,  # noqa : F405
    "Unknown error": 15,
}

for n in errorcode:  # noqa : F405
    posix_error_mappings[os.strerror(n)] = n


def errno_from_msgb(err, map=posix_error_mappings, encoding="utf8", errors="ignore"):
    if err:
        error = err.rsplit(b":", 1)[-1].strip()
    else:
        return default_errorno
    return map.get(error.decode(encoding, errors=errors), default_errorno)


def errno_create_localized_mapping(lang):
    try:
        _ = gettext.translation("libc", languages=[lang])
        _ = _.gettext
    except FileNotFoundError as e:
        warnings.warn(f"Using default mapping for language {lang}: {str(e)}")
        return posix_error_mappings

    localized_map = dict(posix_error_mappings)

    for k, v in posix_error_mappings.items():
        localized_map[_(k)] = v

    return localized_map



import errno
import stat
import time
import os
from myrrh.core.services.system import ExecutionFailureCauseRVal, MOsError, supportedif

from myrrh.framework.mpython._mosfs import AbcOsFs, stat_result
import posixpath

__mlib__ = "OsFs"


def _stat_out_to_struct(out):
    stat_list = [int(v, 0) for v in out.strip().split(b",")]
    stat_list[-3] = float(stat_list[-3])
    stat_list[-2] = float(stat_list[-2])
    stat_list[-1] = float(stat_list[-1])
    return stat_result(
        stat_list,
        {
            "st_atime_ns": stat_list[-3] * 1000000000,
            "st_mtime_ns": stat_list[-2] * 1000000000,
            "st_ctime_ns": stat_list[-1] * 1000000000,
        },
    )


def _statvfs_out_to_struct(out):
    stat_list = [int(v, 0) for v in out.strip().split(b",")]
    return OsFs.statvfs_result(*stat_list)


class OsFs(AbcOsFs):
    normcase = staticmethod(posixpath.normcase)
    splitdrive = staticmethod(posixpath.splitdrive)
    split = staticmethod(posixpath.split)
    samestat = staticmethod(posixpath.samestat)

    _umask = 0o022
    _epochdelta = 3600 if os.name == "nt" else 0

    def link(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None, follow_symlinks=True):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _, err, rval = self.myrrh_os.cmd(
            b"%(ln)s %(follow)s %(src)s %(dst)s",
            follow=b"" if follow_symlinks else b"-n",
            src=self.myrrh_os.sh_escape_bytes(_src),
            dst=self.myrrh_os.sh_escape_bytes(_dst),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, src, error_translate=self._os.default_errno_from_msg).check()

    def chown(self, path, uid, gid, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        _path = self.myrrh_os.p(path, dir_fd=dir_fd)

        if not isinstance(uid, int):
            raise TypeError("invalid user type")
        if not isinstance(gid, int):
            raise TypeError("invalid group type")

        if uid == -1 and gid == -1:
            return

        exe = b"%(chown)s" if uid != -1 else b"%(chgrp)s"
        symlink = b"" if follow_symlinks else b"-h"
        user = b"'%i'" % uid if uid != -1 else b""
        group = b"'%i'" % gid if gid != -1 else b""

        _, err, rval = self.myrrh_os.cmdb(
            b"%s %s %s%s%s %s"
            % (
                exe,
                symlink,
                user,
                b":" if uid != -1 and gid != -1 else b"",
                group,
                self.myrrh_os.sh_escape_bytes(_path),
            )
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, path, error_translate=self._os.default_errno_from_msg).check()

    def chmod(self, path, mode, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        # follow_symlinks is ignore
        _path = self.myrrh_os.p(path, dir_fd=dir_fd)
        mode = stat.S_IMODE(mode)
        _, err, rval = self.myrrh_os.cmdb(
            b"%(chmod)s %(mode)o %(path)s",
            mode=mode,
            path=self.myrrh_os.sh_escape_bytes(_path),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, path, error_translate=self._os.default_errno_from_msg).check()

    def rename(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _, err, rval = self.myrrh_os.cmdb(
            b"%(mv)s %(src)s %(dst)s",
            src=self.myrrh_os.sh_escape_bytes(_src),
            dst=self.myrrh_os.sh_escape_bytes(_dst),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, src, error_translate=self._os.default_errno_from_msg).check()

    def replace(self, src, dst, *args, src_dir_fd=None, dst_dir_fd=None):
        self.rename(src, dst, *args, src_dir_fd=src_dir_fd, dst_dir_fd=dst_dir_fd)

    def listdir(self, path="."):
        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.f(path)

        if not self.isdir(path):
            if not self.exists(path):
                MOsError(self, errno.ENOENT, "No such file or directory", args=(path,)).raised()
            MOsError(self, errno.ENOTDIR, "Not a directory", args=(path,)).raised()

        paths = self.myrrh_os.fs.list(_path)

        return [self.myrrh_os.basename(_cast_(f.strip())) for f, _ in paths]

    def _scandir_list(self, path="."):
        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.p(path)

        files = self.myrrh_os.fs.list(_path)

        result = []
        for file, idx in files:
            if files.readable(idx) or files.writeable(idx) or files.executable(idx):
                fname = self.myrrh_os.basename(file)
                fpath = self.join(_path, fname)
                result.append(self.DirEntry(_cast_(fname), _cast_(fpath), self._stat(fpath), self.lstat))

        return result

    def _stat(self, path, *, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)
        try:
            stat = self.myrrh_os.fs.stat(_path)
        except OSError as err:
            err.filename = path
            raise err

        return stat_result(**stat._asdict())

    def _lstat(self, path, *, dir_fd=None):
        return self._stat(path, dir_fd=dir_fd, follow_symlinks=False)

    @supportedif(lambda self: b"statvfs" in self.myrrh_os.getbinb, "need statvfs tool")
    def _statvfs(self, path):
        _path = self.myrrh_os.f(path)

        command = b" if [ -e %(path)s ] ; then %(stat)s -f -c %%s,%%S,%%b,%%f,%%a,%%c,%%d,%%d,-1,%%l  %(path)s; else ls %(path)s; fi"
        out, err, rval = self.myrrh_os.cmdb(command, path=self.myrrh_os.sh_escape_bytes(_path))
        ExecutionFailureCauseRVal(self, err, rval, 0, path, error_translate=self._os.default_errno_from_msg).check()

        return _statvfs_out_to_struct(out)

    def realpath(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        path = self.myrrh_os.p(path)

        out, err, rval = self.myrrh_os.cmdb(b"%(realpath)s %(path)s", path=self.myrrh_os.sh_escape_bytes(path))
        ExecutionFailureCauseRVal(self, err, rval, 0, error_translate=self._os.default_errno_from_msg).check()

        return _cast_(out.strip())

    def ismount(self, path):
        path = self.myrrh_os.p(path)
        # copy from posixpath
        if self.islink(path):
            # A symlink can never be a mount point
            return False
        try:
            s1 = self.lstat(path)
            s2 = self.lstat(self.join(path, self.myrrh_os.pardirb))
        except ExecutionFailureCauseRVal:
            return False  # It doesn't exist -- so not a mount point:-)
        dev1 = s1.st_dev
        dev2 = s2.st_dev
        if dev1 != dev2:
            return True  # path/.. on a different device as path
        ino1 = s1.st_ino
        ino2 = s2.st_ino
        if ino1 == ino2:
            return True  # path/.. is the same i-node as path
        return False

    def __utime_toolbox(self, bpath, times, follow_symlinks):
        if times is None:
            _, err, rval = self.myrrh_os.cmdb(
                b"%(touch)s %(symlink)s -a -m %(path)s",
                symlink="-l" if follow_symlinks else b"",
                path=self.myrrh_os.sh_escape_bytes(bpath),
            )

        else:
            atime, mtime = times
            atime = self.myrrh_os.shencode(time.strftime("%Y%m%d.%H%M%S", time.localtime(atime)))
            mtime = self.myrrh_os.shencode(time.strftime("%Y%m%d.%H%M%S", time.localtime(mtime)))

            _, err, rval = self.myrrh_os.cmdb(
                b"%(touch)s %(follow)s -a -t %(atime)s %(path)s && %(touch)s %(follow)s -m -t %(mtime)s %(path)s",
                follow=b"-l" if follow_symlinks else b"",
                atime=atime,
                mtime=mtime,
                path=self.myrrh_os.sh_escape_bytes(bpath),
            )
        return err, rval

    def __utime_toybox(self, bpath, times, follow_symlinks):
        if times is None:
            _, err, rval = self.myrrh_os.cmdb(
                b"%(touch)s %(follow)s -c -a %(path)s",
                follow=b"" if follow_symlinks else b"-h",
                path=self.myrrh_os.sh_escape_bytes(bpath),
            )

        else:
            atime, mtime = times
            atime = atime and self.myrrh_os.shencode(time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(atime)))
            mtime = mtime and self.myrrh_os.shencode(time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(mtime)))

            cmd = b" && ".join(
                filter(
                    None,
                    [
                        atime and b"%(touch)s %(follow)s -c -a -d %(atime)s %(path)s",
                        mtime and b"%(touch)s %(follow)s -m -d %(mtime)s %(path)s",
                    ],
                )
            )

            _, err, rval = self.myrrh_os.cmdb(
                cmd,
                follow=b"" if follow_symlinks else b"-h",
                atime=atime,
                mtime=mtime,
                path=self.myrrh_os.sh_escape_bytes(bpath),
            )
        return err, rval

    def _utime(self, path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        if times or ns:
            atime, mtime = times or (t * 1e-9 for t in ns)
            atime, mtime = (
                atime and atime + self._epochdelta,
                mtime and mtime + self._epochdelta,
            )

        if self.myrrh_os.getbinb[b"touch"].startswith(b"/system/bin/toolbox"):
            _utime_cmd = self.__utime_toolbox
        else:
            _utime_cmd = self.__utime_toybox

        err, rval = _utime_cmd(_path, (atime, mtime), follow_symlinks=follow_symlinks)

        ExecutionFailureCauseRVal(self, err, rval, 0, path, error_translate=self._os.default_errno_from_msg).check()

    def umask(self, mask):
        last_mask = self._umask
        self._umask = mask
        return last_mask

    def truncate(self, path, length):
        _path = self.myrrh_os.p(path)
        if length:
            _, err, rval = self.myrrh_os.cmdb(
                b"%(dd)s  bs=%(length)d count=1 skip=2 seek=1 of=%(path)s if=%(path)s\n",
                length=length,
                path=self.myrrh_os.sh_escape_bytes(_path),
            )
        else:
            _, err, rval = self.myrrh_os.cmdb(
                b'%(sh)s -c "exec 1>%(path)s"',
                path=self.myrrh_os.sh_escape_bytes(_path),
            )

        ExecutionFailureCauseRVal(self, err, rval, 0, path, error_translate=self._os.default_errno_from_msg).check()

    def samefile(self, path1, path2):
        stat1 = self.stat(path1)
        stat2 = self.stat(path2)

        return self.samestat(stat1, stat2) and (self.normcase(self.abspath(path1)) == self.normcase(self.abspath(path2)))



# -*- coding: utf-8 -*-

import os
import errno
import tempfile
import threading

from myrrh.core.interfaces import abstractmethod
from myrrh.core.services import PID, cfg_init
from myrrh.core.services.system import AbcRuntime, FileException, _mlib_
from ..mpython import _mosfs

__mlib__ = "AbcAdvFs"


class AdvFsFile:
    def __init__(self, arg):
        self.arg = arg
        self._file = None

    def __enter__(self):
        self._file = self.open()
        return self._file

    def __exit__(self, *a, **kwa):
        self.close()

    def close(self):
        if self._file:
            self._file.close()
            self._file = None


class AdvFsFileW(AdvFsFile):
    def open(self):
        if self.arg is None:
            return tempfile.SpooledTemporaryFile()

        if hasattr(self.arg, "write"):
            return self.arg

        return open(self.arg, "wb")


class AdvFsFileR(AdvFsFile):
    def open(self):
        if hasattr(self.arg, "read"):
            return self.arg

        return open(self.arg, "rb")


class AdvFsFileGet(AdvFsFile):
    def __init__(self, system: AbcRuntime, path, size, header=b""):
        super().__init__((system, path, size))

        self.system = system
        self.path = path
        self._thread = threading.Thread(target=self._run)

        self.rfile, self.wfile = system.myrrh_syscall.open_pipe()
        self.rfile, self.wfile = system.myrrh_syscall.gethandle(self.rfile), system.myrrh_syscall.gethandle(self.wfile)  # type: ignore[assignment]

        self.header = header

    def _run(self):
        if self.header:
            sz = 0
            while sz < len(self.header):
                sz += self.wfile.write(self.header[sz:])
        try:
            self.system.myrrh_syscall.stream_in(self.path, self.wfile)
        finally:
            self.wfile.close()

    def open(self):
        self._thread.start()
        return self.rfile

    def close(self):
        self._thread.join()
        super().close()


class AbcAdvFs(AbcRuntime):
    __frameworkpath__ = "mfs.advfs"

    CHUNK_SZ = cfg_init("advfs_file_chunk_size", 1024 * 500, section="myrrh.framework.mfs")
    COPY_BUFSIZE = cfg_init("advfs_copy_buffer_size", 1024 * 64, section="myrrh.framework.mfs")

    @property
    def os(self):
        return _mlib_(_mosfs)(self)

    @abstractmethod
    def _mkdirs(self, dirs: list):
        pass

    @abstractmethod
    def _chunk_header(self, file_descs):
        pass

    @abstractmethod
    def _unchunk(self, path):
        pass

    @abstractmethod
    def _write_chunk(self, path, file_descs):
        pass

    @abstractmethod
    def _scanfiles(self, files):
        """
        return files(path, size)
        """

    @abstractmethod
    def _scantree(self, path):
        """
        return dirs(path), files(path, size)
        """

    def scanfiles(self, files):
        """
        return files(path, size)
        """
        _cast_ = (self.myrrh_os.fdcast(f) for f in files)
        files = (self.myrrh_os.p(f) for f in files)
        files = self._scanfiles(files)

        return [_c_(f) for _c_, (f, _) in zip(_cast_, files)], [sz for (f, sz) in files]

    def scantree(self, path):
        """
        return dirs(path), files(path, size)
        """
        _cast_ = self.myrrh_os.fdcast(path)
        path = self.myrrh_os.p(path)

        root = path if path.endswith(self.myrrh_os.sepb) else path + self.myrrh_os.sepb
        root = self.myrrh_os.getpathb(root)

        dirs, files = self._scantree(path)

        dirs = (d[len(root) + 1 :] if d.startswith(root) else d for d in dirs)

        _files = []
        _szs = []
        for f, sz in files:
            _files.append(f[len(root) + 1 :] if f.startswith(root) else f)
            _szs.append(sz)

        return [_cast_(d) for d in dirs], [_cast_(f) for f in _files], _szs

    def local_scantree(self, path):
        dirs = []
        files = []
        sizes = []

        path = path if path[-1] == os.sep else path + os.sep

        def _scan(path):
            with os.scandir(path) as scan:
                for ent in scan:
                    yield ent
                    if ent.is_dir():
                        yield from _scan(ent.path)

        for ent in _scan(path):
            ent_path = ent.path[len(path) :] if ent.path.startswith(path) else ent.path
            if ent.is_dir():
                dirs.append(ent_path)
            else:
                files.append(ent_path)
                sizes.append(os.path.getsize(os.path.join(path, ent_path)))

        return dirs, files, sizes

    def trpath(self, path):
        def _trpath(path):
            _cast_ = self.myrrh_os.fdcast(path)
            path = self.myrrh_os.p(path)

            if self.myrrh_os._sepb_ == b"\\":
                path = path.replace(b"/", self.myrrh_os._sepb_)
            else:
                path = path.replace(b"\\", self.myrrh_os._sepb_)

            return _cast_(path)

        if isinstance(path, list):
            return list(_trpath(p) for p in path)

        if isinstance(path, tuple):
            return tuple(_trpath(p) for p in path)

        return _trpath(path)

    def _local_write_chunk(self, stream, file_descs):
        stream.write(self._chunk_header(file_descs))

        for src, _, _ in file_descs:
            with AdvFsFileR(src) as f:
                self.copyfileobj(f, stream)

    def _chunk_file_name(self, chunk_nb=None, chunk_name=None):
        if chunk_nb is None:
            return b"%d%d%s" % (PID, id(self), self.myrrh_os.cfg.uuid.encode())

        if chunk_name:
            return b"%s%d.chunk" % (chunk_name, chunk_nb)

        return b"%d%d%s%d.chunk" % (
            PID,
            id(self),
            self.myrrh_os.cfg.uuid.encode(),
            chunk_nb,
        )

    def _local_makechunks(self, srcs, dests, sizes, chunk_size, chunk_name=None):
        # return src, dest, ischunk, file_descs
        nb_chunk = 0
        file_descs = []
        total_sz = 0

        tempdir = self.myrrh_os.tmpdirb

        for file_desc in zip(srcs, dests, sizes):
            src, dest, size = file_desc

            if size is None:
                FileException(self, errno=errno.ENOENT, filename=src).raised()

            if size > chunk_size:
                yield src, dest, False, [file_desc]

            elif total_sz + size < chunk_size:
                file_descs.append(file_desc)

            else:
                chunkpath = self.myrrh_os.joinpath(tempdir, self._chunk_file_name(nb_chunk, chunk_name=chunk_name))
                try:
                    stf = tempfile.SpooledTemporaryFile()
                    self._local_write_chunk(stf, file_descs)
                    stf.seek(0)
                    yield stf, chunkpath, True, file_descs
                    nb_chunk += 1
                finally:
                    stf.close()

                file_descs = [file_desc]

        # only one file, direct download

        if len(file_descs) == 1:
            src, dest, size = file_descs[0]
            yield src, dest, False, file_descs

        elif file_descs:
            try:
                chunkpath = self.myrrh_os.joinpath(tempdir, self._chunk_file_name(nb_chunk, chunk_name=chunk_name))
                stf = tempfile.SpooledTemporaryFile()
                self._local_write_chunk(stf, file_descs)
                stf.seek(0)
                yield stf, chunkpath, True, file_descs
            finally:
                stf.close()

    def _makechunks(self, srcs, dests, sizes, chunk_size, chunk_name=None):
        # return src, dest, ischunk, file_descs
        nb_chunk = 0
        file_descs = []
        total_sz = 0

        tempdir = self.myrrh_os.tmpdirb

        for file_desc in zip(srcs, dests, sizes):
            src, dest, size = file_desc

            if size is None:
                FileException(self, errno=errno.ENOENT, filename=src).raised()

            if size > chunk_size:
                yield src, dest, False, [file_desc]

            elif total_sz + file_desc[2] < chunk_size:
                file_descs.append(file_desc)
            else:
                chunkpath = self.myrrh_os.joinpath(tempdir, self._chunk_file_name(nb_chunk, chunk_name=chunk_name))
                self._write_chunk(chunkpath, file_descs)
                yield chunkpath, None, True, file_descs
                nb_chunk += 1
                file_descs = [file_desc]

        # only one file, direct transfer

        if len(file_descs) == 1:
            src, dest, size = file_descs[0]
            yield src, dest, False, file_descs

        elif file_descs:
            chunkpath = self.myrrh_os.joinpath(tempdir, self._chunk_file_name(nb_chunk, chunk_name=chunk_name))
            self._write_chunk(chunkpath, file_descs)
            yield chunkpath, None, True, file_descs

    def copyfileobj(self, fsrc, fdst, sz=0, length=0):
        """copy data from file-like object fsrc to file-like object fdst"""
        # Localize variable access to minimize overhead.
        if not length:
            length = self.COPY_BUFSIZE
        rd_sz = 0
        while not sz or rd_sz < sz:
            to_rd = min(sz - rd_sz, length) if sz else length
            buf = fsrc.read(to_rd)
            if not buf:
                break
            fdst.write(buf)
            rd_sz += len(buf)

        return rd_sz

    def mkdirs(self, dirs):
        dirs.sort()

        filtered_dirs = []
        for i in range(0, len(dirs) - 1):
            if not dirs[i] in dirs[i + 1]:
                filtered_dirs.append(self.myrrh_os.sh_escape_bytes(self.myrrh_os.p(dirs[i])))
        if dirs:
            filtered_dirs.append(self.myrrh_os.sh_escape_bytes(self.myrrh_os.p(dirs[-1])))

        self._mkdirs(filtered_dirs)

    def getfiles(
        self,
        src_files,
        dest_files=[],
        *,
        sizes=None,
        chunk_size=CHUNK_SZ,
        ignore_overwrite=False,
    ):
        result = []

        if not dest_files:
            dest_files = [os.path.join(os.getcwd(), self.basename(f)) for f in src_files]

        if len(src_files) != len(dest_files):
            raise ValueError("number of elements in source file list and destination path list must be equal")

        if not ignore_overwrite:
            for f in dest_files:
                if dest_files.count(f) != 1:
                    raise ValueError(f"destination {f} file already exists")

        if not chunk_size:
            for src, dest in zip(src_files, dest_files):
                src = self.myrrh_os.p(src)

                with AdvFsFileW(dest) as f:
                    self.myrrh_syscall.stream_in(src, f)
                    result.append(dest)

            return result

        if not sizes:
            _, sizes = self.scanfiles(src_files)

        for src, dest, ischunk, file_descs in self._makechunks(src_files, dest_files, sizes, chunk_size):
            with AdvFsFileW(dest) as stream:
                self.myrrh_syscall.stream_in(os.fsencode(src), stream)

                if ischunk:
                    stream.seek(0)
                    for _, dest, sz in file_descs:
                        with AdvFsFileW(dest) as f:
                            self.copyfileobj(stream, f, sz)

            result.extend(d for _, d, _ in file_descs)

        return result

    def getfile(self, src, dest=""):
        dests = [] if not dest else [dest]
        return self.getfiles([src], dests, chunk_size=0)

    def getdir(self, src, dest="", *, chunk_size=CHUNK_SZ):
        if not dest:
            dest = os.getcwd()

        dirs, files, sizes = self.scantree(src)

        if os.path.isdir(dest) or dest.endswith(os.sep):
            dest = os.path.join(dest, self.myrrh_os.basename(src))
            os.makedirs(dest, exist_ok=True)

        for d in dirs:
            dir = os.path.join(dest, *d.split(self.myrrh_os.sepb.decode()))
            os.makedirs(dir, exist_ok=True)

        return dirs, self.getfiles(
            [self.myrrh_os.joinpath(src, f) for f in files],
            [os.path.join(dest, *f.split(self.myrrh_os.sepb.decode())) for f in files],
            sizes=sizes,
            chunk_size=chunk_size,
        )

    def pushfiles(self, src_files, dest_files=[], *, sizes=None, chunk_size=CHUNK_SZ):
        if not dest_files:
            dest_files = [self.myrrh_os.joinpath(self.myrrh_os._getcwdb_(), os.basename(f)) for f in src_files]

        if len(src_files) != len(dest_files):
            raise ValueError("number of elements in source file list and destination path list must be equal")

        result = []
        if not chunk_size:
            for src, dest in zip(src_files, dest_files):
                with AdvFsFileR(src) as f:
                    self.myrrh_syscall.stream_out(self.myrrh_os.fsencode(dest), f)
                    result.append(dest)

            return result

        if not sizes:
            sizes = []
            for f in src_files:
                sizes.append(os.path.getsize(f))

        for src, dest, merged, file_descs in self._local_makechunks(src_files, dest_files, sizes, chunk_size=chunk_size):
            with AdvFsFileR(src) as stream:
                self.myrrh_syscall.stream_out(os.fsencode(dest), stream)
            if merged:
                self._unchunk(dest)

            result.extend(d for _, d, _ in file_descs)

        return result

    def pushfile(self, src, dest=""):
        if self.myrrh_os.fs.is_container(self.myrrh_os.fsencode(dest)):
            dest = self.myrrh_os.joinpath(dest, os.path.basename(src))

        dests = [] if not dest else [dest]

        return self.pushfiles([src], dests, chunk_size=0)

    def pushdir(self, src, dest="", *, chunk_size=CHUNK_SZ):
        if not dest:
            dest = self.myrrh_os.fsdecode(self.myrrh_os._getcwdb_())

        dirs, files, sizes = self.local_scantree(src)
        dirs = [self.myrrh_os.joinpath(*d.split(os.sep)) for d in dirs]

        if dest.endswith(self.myrrh_os.sepb.decode()) or self.myrrh_os.fs.is_container(self.myrrh_os.p(dest)):
            dest = self.myrrh_os.joinpath(dest, os.path.basename(src))

        self.mkdirs([self.myrrh_os.joinpath(dest, d) for d in dirs])

        files = self.pushfiles(
            [os.path.join(src, f) for f in files],
            [self.myrrh_os.joinpath(dest, *f.split(os.sep)) for f in files],
            sizes=sizes,
            chunk_size=chunk_size,
        )

        return dirs, files

    def trmod(self, srcpath, destpath, mode=-1):
        self.myrrh_os.p = self.myrrh_os.p(destpath)
        srcpath = os.fsencode(srcpath)

        fmode = mode if mode != -1 else ("%o" % os.stat(srcpath).st_mode)[-3:] if os.name == self.myrrh_os.cfg.os else self.default_mode

        if fmode != -1:
            self.chmod(fmode, destpath)

    def transferfiles(self, entity, src_files, dest_files=[], *, sizes=None, chunk_size=CHUNK_SZ):
        result = []

        entity = AbcAdvFs(entity.system)

        if not dest_files:
            dest_files = [entity.myrrh_os.basename(f) for f in src_files]

        if len(src_files) != len(dest_files):
            raise ValueError("number of elements in source file list and destination path list must be equal")

        if sizes is None:
            _, sizes = entity.scanfiles(src_files)

        if len(src_files) != len(sizes):
            raise ValueError("number of elements in source file list and associated size list differs")

        dest_files = [self.trpath(d) for d in dest_files]

        if not chunk_size:
            for src, dest, sz in zip(src_files, dest_files, sizes):
                if sz is None:
                    FileException(entity, errno=errno.ENOENT, filename=src).raised()

                with AdvFsFileGet(entity, entity.myrrh_os.p(src), sz) as f:
                    self.myrrh_syscall.stream_out(self.myrrh_os.fsencode(dest), f)
                result.append(dest)

            return result

        target_tempdir = self.myrrh_os.tmpdirb
        for src, dest, ischunk, file_descs in entity._makechunks(src_files, dest_files, sizes, chunk_size, chunk_name=self._chunk_file_name()):
            header = self._chunk_header(file_descs) if ischunk else b""
            size = sum(sz for _, _, sz in file_descs)

            with AdvFsFileGet(entity, entity.myrrh_os.p(src), size, header) as stream:
                if dest is None:
                    dest = self.myrrh_os.joinpath(
                        target_tempdir,
                        entity.myrrh_os.p(entity.myrrh_os.basename(src) + b"_"),
                    )
                self.myrrh_syscall.stream_out(self.myrrh_os.fsencode(dest), stream)

                if ischunk:
                    self._unchunk(dest)

            if ischunk:
                entity.rm(src)

            result.extend(d for _, d, _ in file_descs)

        return result

    def transferfile(self, entity, src, dest=""):
        if dest and self.myrrh_os.fs.is_container(dest):
            dest = self.myrrh_os.joinpath(dest, entity.runtime.myrrh_os.basename(src))

        dests = [] if not dest else [dest]
        return self.transferfiles(entity, [src], dests, chunk_size=0)

    def transferdir(self, entity, src, dest="", *, chunk_size=CHUNK_SZ):
        e_advfs = AbcAdvFs(entity.system)

        _cast_ = e_advfs.myrrh_os.fdcast(src)

        src = e_advfs.myrrh_os.p(src)
        dest = self.myrrh_os.p(dest)

        if not dest:
            dest = self.myrrh_os.getpathb()

        if self.myrrh_os.fs.is_container(dest) or dest.endswith(self.myrrh_os.sepb):
            dest = self.myrrh_os.joinpath(dest, e_advfs.myrrh_os.fsencode(e_advfs.myrrh_os.basename(src)))

        dirs, files, sizes = e_advfs.scantree(src)

        self.mkdirs(
            list(
                map(
                    lambda d: self.myrrh_os.joinpath(dest, self.myrrh_os.fsencode(self.trpath(d))),
                    dirs,
                )
            )
        )
        files = self.transferfiles(
            entity,
            [e_advfs.myrrh_os.joinpath(src, f) for f in files],
            [self.myrrh_os.joinpath(dest, self.myrrh_os.fsencode(f)) for f in files],
            sizes=sizes,
            chunk_size=chunk_size,
        )

        return [_cast_(d) for d in dirs], [_cast_(f) for f in files]

    def transfer(self, entity, src, dest="", *, chunk_size=CHUNK_SZ):
        if entity.runtime.myrrh_os.fs.is_container(src):
            return self.transferdir(entity, src, dest, chunk_size=chunk_size)

        return [], self.transferfile(entity, src, dest=dest)

    def copy(self, srcs, dest):
        """
        copy dirs or files entries to other location on the entity

        if dest is a directory src is copy inside dest
        """
        dest = self.myrrh_os.p(dest)

        srcs = [self.myrrh_os.p(src) for src in srcs] if isinstance(srcs, (list, tuple)) else [self.myrrh_os.p(srcs)]

        self._localtransfer(srcs, dest)

    @abstractmethod
    def _localtransfer(self, src, dest):
        """
        copy dirs or files entries to other location on the entity

        if dest is a directory src is copy inside dest
        """

    def move(self, srcs, dest):
        """
        move dirs or files entries to other location on the entity

        if dest is a directory src is copy inside dest
        """
        dest = self.myrrh_os.p(dest)

        srcs = [self.myrrh_os.p(src) for src in srcs] if isinstance(srcs, (list, tuple)) else [self.myrrh_os.p(srcs)]

        self._localmove(srcs, dest)

    @abstractmethod
    def _localmove(self, src, dest):
        ...

    def rm(self, srcs):
        """
        remove complete directory trees or files entries
        """
        srcs = [self.myrrh_os.p(src) for src in srcs] if isinstance(srcs, (list, tuple)) else [self.myrrh_os.p(srcs)]

        self._localremove(srcs)

    @abstractmethod
    def _localremove(self, srcs):
        ...


AdvFs = AbcAdvFs



import click
import sys
import os

import myrrh.factory
import pprint

__all__ = ["json"]


class _F:
    @staticmethod
    def dump_schema(factory):
        return factory.schema(indent=2)

    @staticmethod
    def dump_json(factory):
        if isinstance(factory, myrrh.factory.Assembly):
            return factory.json(indent=2)

    @staticmethod
    def dump_obj(factory):
        if isinstance(factory, myrrh.factory.Assembly):
            return factory.pformat()

    @staticmethod
    def dump_dict(factory):
        if isinstance(factory, myrrh.factory.Assembly):
            return pprint.pformat(factory.dict())


@click.option("-o", "--dump-obj", "dump", flag_value="dump_obj", default=False)
@click.option("-d", "--dump-dict", "dump", flag_value="dump_dict", default=False)
@click.option("-s", "--dump-schema", "dump", flag_value="dump_schema", default=True)
@click.option("-j", "--dump-json", "dump", flag_value="dump_json", default=False)
@click.option("-l", "--load", type=str, default="")
@click.option("-e", "--export", type=str, default="")
@click.option("-y", "--yes", is_flag=True, default=False)
def json(dump, load, export, yes):
    try:
        Factory = myrrh.factory.Assembly

        if load:
            Factory = myrrh.factory.Assembly.fromFile(load)

        try:
            CONT = True

            if export == "-":
                export = load

            CONT = not export or not os.path.exists(export) or yes or click.confirm(f"{export} file exist, overwrite?")

            if CONT and export:
                export = open(export, "w")

            if CONT and dump:
                click.echo(getattr(_F, dump)(Factory), file=export or sys.stdout)

        finally:
            if CONT and export:
                export.close()

    except Exception as e:
        raise
        click.echo(e, err=True)


if __name__ == "__main__":
    json("dump_schema", "", "", False)



import click
import pprint


class _F:
    from myrrh.core.services import cfg_get, cfg_set, cfg_del  # type: ignore[misc]

    @staticmethod
    def dump(key, value, section, overwrite):
        return pprint.pformat(_F.cfg_get(key, section=section))

    @staticmethod
    def rm(key, value, section, overwrite):
        to_delete = _F.dump(key, value=value, section=section, overwrite=overwrite)
        to_delete = to_delete[:200] + " ..." if len(to_delete) >= 200 else to_delete
        if not overwrite:
            click.echo("the following code will be deleted from the myrrh config")
            click.echo("--")
            click.echo(to_delete)
            click.echo("--")
            if not click.confirm("Do you want to continue?"):
                return "Aborted !"

        _F.cfg_del(key, section=section)
        return f"{key}={to_delete} deleted"


__all__ = ["cfg"]


def _confirm_cbk(ctx, param, value):
    if value == "rm":
        _d = {p.name: p for p in ctx.command.params}
        print(dir(_d.get("section")))
        print(_d)
        if not click.confirm("Do you want to continue?"):
            ctx.abort()

    return param


@click.option("--dump", "action", flag_value="dump", default=True)
@click.option("--set", "action", flag_value="cfg_set", default=False)
@click.option("--rm", "action", flag_value="rm", default=False)
@click.option("-k", "--key", type=str, default="")
@click.option("-s", "--section", type=str, default="")
@click.option("-f", "--force", is_flag=True, default=False)
@click.option("-v", "--value", default="")
def cfg(action, key, section, force, value):
    if action:
        click.echo(getattr(_F, action)(key, value=value, section=section, overwrite=force))


if __name__ == "__main__":
    click.command(cfg)()



from myrrh.framework.mpython.mtest.msupport.mos_helper import AbcOsHelper
from myrrh.framework.mpython.mtest.msupport.mimport_helper import AbcImportHelper
from myrrh.framework.mpython.mtest.msupport.mscript_helper import AbcScriptHelper

os_helper: AbcOsHelper
import_helper: AbcImportHelper
script_helper: AbcScriptHelper



import typing

from io import TextIOWrapper

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcBz2"


class _interface(ABC):
    import bz2 as local_bz2

    @property
    @abstractmethod
    def BZ2Compressor(self) -> local_bz2.BZ2Compressor:
        ...

    @property
    @abstractmethod
    def BZ2Decompressor(self) -> local_bz2.BZ2Decompressor:
        ...

    @property
    @abstractmethod
    def BZ2File(self) -> local_bz2.BZ2File:
        ...

    @property
    @abstractmethod
    def _MODE_CLOSED(self):
        return self.local_bz2._MODE_CLOSED

    @property
    @abstractmethod
    def _MODE_READ(self):
        return self.local_bz2._MODE_READ

    @property
    @abstractmethod
    def _MODE_WRITE(self):
        return self.local_bz2._MODE_WRITE

    @abstractmethod
    def open(
        self,
        filename,
        mode="rb",
        compresslevel=9,
        encoding=None,
        errors=None,
        newline=None,
    ) -> TextIOWrapper | local_bz2.BZ2File:
        ...

    @abstractmethod
    def compress(self, data, compresslevel=9) -> typing.Any:
        ...

    @abstractmethod
    def decompress(self, data) -> typing.Any:
        ...


class AbcBz2(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mbz2"

    __all__ = [
        "BZ2File",
        "BZ2Compressor",
        "BZ2Decompressor",
        "open",
        "compress",
        "decompress",
    ]

    __doc__ = _interface.local_bz2.__doc__

    __delegated__ = {_interface: _interface.local_bz2}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_bz2, self)

        self.__delegate__(_interface, mod)



import click
from click.core import Group
from .cmd import myrrhc_cmds

__all__ = ["__doc__", "HelpFormatter"]


class HelpFormatter(click.HelpFormatter):
    """This class helps with formatting text-based help pages.  It's
    usually just needed for very special internal cases, but it's also
    exposed so that developers can write their own fancy outputs.

    At present, it always writes into memory.

    :param indent_increment: the additional increment for each level.
    :param width: the width for the text.  This defaults to the terminal
                  width clamped to a maximum of 78.
    """

    def write_usage(self, prog: str, args: str = "", prefix=None) -> None:
        """Writes a usage line into the buffer.

        :param prog: the program name.
        :param args: whitespace separated list of arguments.
        :param prefix: The prefix for the first line. Defaults to
            ``"Usage: "``.
        """
        if prefix is None:
            prefix = "Usage"

        # The prefix is too long, put the arguments on the next line.
        self.write_heading(prefix)
        self.write(f"{prog} ")
        self.write(args)
        self.write("\n")

    def write_heading(self, heading: str) -> None:
        self.write(f"{'':>{self.current_indent}}{heading}\n---\n")

    def write_dl(
        self,
        rows,
        col_max: int = 30,
        col_spacing: int = 2,
    ) -> None:
        """Writes a definition list into the buffer.  This is how options
        and commands are usually formatted.

        :param rows: a list of two item tuples for the terms and values.
        :param col_max: the maximum width of the first column.
        :param col_spacing: the number of spaces between the first and
                            second column.
        """
        rows = list(rows)
        widths = click.formatting.measure_table(rows)
        if len(widths) != 2:
            raise TypeError("Expected two columns for definition list")

        self.write('<table border="1" cellpadding="8">')
        for first, second in click.formatting.iter_rows(rows, len(widths)):
            self.write("<tr>")

            if not first.startswith("-"):
                self.write(f'<th><a href="_main_/{first}.html">{first}</a></th>')
            else:
                self.write(f"<th>{first}</th>")

            if not second:
                self.write("</tr>")
                continue
            lines = second.splitlines()

            if lines:
                self.write(f"<td>{lines[0]}</td>")

                for line in lines[1:]:
                    self.write(f"<td>{line}</td>")

            self.write("</tr>")
        self.write("</table>")


def __doc__():  # type: ignore[no-redef]
    ctx = myrrhc_cmds.make_context("myrrhc", ["$"], None, resilient_parsing=True)
    formatter = HelpFormatter()
    ctx.command.format_help(ctx, formatter)

    return formatter.getvalue()


_alias = {"commands": "myrrh.tools.myrrhc"}


def alias():
    _alias.update({cmdname: f"commands#{cmdname}" for cmdname in myrrhc_cmds.commands.keys() if not cmdname.startswith("-")})

    return _alias


def register_alias(alias_dict):
    _alias.update(alias_dict)


def __docgen__(cmdname=None, group=myrrhc_cmds):
    ctx = group.make_context("myrrhc", ["$"], None, resilient_parsing=True)
    cmdnames = [cmdname] if cmdname else [cmdname for cmdname in group.commands.keys() if not cmdname.startswith("-")]

    result = dict()
    for cname in cmdnames:
        cmd = group.get_command(ctx, cname)
        sub_ctx = cmd.make_context(cname, [cname], ctx, resilient_parsing=True)
        formatter = HelpFormatter()
        sub_ctx.command.format_help(sub_ctx, formatter)
        result[cname] = formatter.getvalue()

        if isinstance(cmd, Group):
            for k, txt in __docgen__(group=cmd).items():
                result[f"{cname}.{k}"] = txt

    return result



import threading
import time
import typing

from concurrent.futures import Future, CancelledError, InvalidStateError

from ...services import log
from ...interfaces import ITask, IRuntimeTaskManager
from ....utils import mtimer


__all__ = ("RuntimeTask",)

PENDING = "PENDING"
RUNNING = "RUNNING"
IDLE = "IDLE"
CANCELLED = "CANCELLED"
CANCELLED_AND_NOTIFIED = "CANCELLED_AND_NOTIFIED"
FINISHED = "FINISHED"


# copy of "Future" source code in concurrent.futures
# adapted to support idle state
# (consult the PSF for licence agreement)
class TaskFuture(Future):
    def __enter__(self):
        return self._condition.__enter__()

    def __exit__(self, *args):
        return self._condition.__exit__(*args)

    def idle(self, timeout=None):
        try:
            with self._condition:
                if self._state in [CANCELLED, CANCELLED_AND_NOTIFIED]:
                    raise CancelledError()
                elif self._state in [IDLE, FINISHED]:
                    return super().result()

                self._condition.wait(timeout)

                if self._state in [CANCELLED, CANCELLED_AND_NOTIFIED]:
                    raise CancelledError()
                elif self._state in [IDLE, FINISHED]:
                    return super().result()
                else:
                    raise TimeoutError()
        finally:
            self = None

    def result(self, timeout=None):
        try:
            with self._condition:
                if self._state in [CANCELLED, CANCELLED_AND_NOTIFIED]:
                    raise CancelledError()
                elif self._state == FINISHED:
                    return super().result()

                timer = mtimer.MTimer(timeout=timeout)
                while self._state in [RUNNING, IDLE, PENDING]:
                    self._condition.wait(timer.timeleft)

                if self._state in [CANCELLED, CANCELLED_AND_NOTIFIED]:
                    raise CancelledError()
                elif self._state == FINISHED:
                    return super().result()
                else:
                    raise TimeoutError()
        finally:
            self = None

    def running(self):
        with self._condition:
            return self._state in [RUNNING, IDLE]

    def idled(self):
        with self._condition:
            return self._state == IDLE

    def set_idle(self, result):
        with self._condition:
            if self._state in {CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED}:
                raise InvalidStateError("{}: {!r}".format(self._state, self))
            self._result = result
            self._state = IDLE
            for waiter in self._waiters:
                waiter.add_result(self)
            self._condition.notify_all()
        self._invoke_callbacks()

    def notify_all(self):
        with self._condition:
            for waiter in self._waiters:
                waiter.add_result(self)
            self._condition.notify_all()
        self._invoke_callbacks()

    def wait(self, timeout):
        with self._condition:
            if not self._condition.wait(timeout):
                raise TimeoutError()


class RuntimeTask:
    def __init__(self, task: ITask, manager: IRuntimeTaskManager):
        self.future = TaskFuture()
        self.runtime_time = 0
        self.manager = manager
        self.task = task

    def __del__(self):
        try:
            self.future.set_result(None)
        except InvalidStateError:
            pass

    def run(self):
        ct = threading.current_thread()
        self.manager
        ct.name = f"{self.manager.name}{str(self.task)}"

        if not self.future.idled:
            self.future.set_running_or_notify_cancel()

        st_time = time.monotonic()

        try:
            value = self.task.task()
        except Exception as exc:
            log.debug(f"myrrh task {ct.name} ended with exception: {exc}")

            endtime = time.monotonic() - st_time
            self.runtime_time += endtime

            try:
                self.future.set_exception(exc)
            except InvalidStateError:
                pass

            self = None

            ct.name = f"{ct.name}(exception)"

        else:
            endtime = time.monotonic() - st_time
            self.runtime_time += endtime

            with self.future:
                result = self.task.terminated()
                if result is not None:
                    self.future.set_result(result)
                    ct.name = f"{ct.name}(terminated)"
                else:
                    self.future.set_idle(value)
                    ct.name = f"{ct.name}(idle)"

            try:
                if self.future.idled():
                    try:
                        self.manager.submit(self)
                    except RuntimeError as exc:
                        log.debug(f"myrrh task {ct.name} can not leave idle state: {exc}")

                        self.future.set_exception(exc)

            except InvalidStateError:
                pass

    def join(self, timeout: float | None = None) -> typing.Any:
        return self.future.result(timeout)



import asyncio
import sys

from myrrh.tools.myrrhc.exceptions import Reboot

from myrrh.tools.myrrhc import prompt
from myrrh.tools.myrrhc_ext import myrrhc_cmds


def getsession():
    return Cli._session


class Cli:
    _session: prompt.PromptSession | None = None

    @classmethod
    def run(cls):
        try:
            with myrrhc_cmds.make_context(sys.argv[0], list()) as ctx:
                loop = asyncio.get_event_loop()

                Cli._session = prompt.PromptSession(ctx)

                if len([v for v in filter(lambda s: s not in ("-v", "--verbose"), sys.argv)]) == 1:
                    loop.run_until_complete(Cli._session.run())
                else:
                    loop.run_until_complete(Cli._session.command_call(" ".join(sys.argv[1:])))
        except EOFError:
            pass
        except Reboot:
            sys.exit(2)

        sys.exit(0)



from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "Abc_Compression"


class _interface(ABC):
    import _compression as local__compression

    @property
    @abstractmethod
    def BaseStream(self) -> local__compression.BaseStream:
        ...

    @property
    @abstractmethod
    def DecompressReader(self) -> local__compression.DecompressReader:
        ...

    @property
    @abstractmethod
    def BUFFER_SIZE(self):
        return self.local__compression.BUFFER_SIZE


class Abc_Compression(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython._mcompression"

    __doc__ = _interface.local__compression.__doc__

    __delegated__ = {_interface: _interface.local__compression}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local__compression, self)

        self.__delegate__(_interface, mod)



import bmy
import os

from myrrh.core.services import cfg_init

__all__ = ("setUp", "setUpModule", "tearDownModule")

test_path = cfg_init("testdir", ".myrrh_tests", section="tests")

_eid: str | None = None
_name: str | None = None
_local_prev_cwd: str | None = None


def setUpModule():
    init(_name, eid=_eid)


def tearDownModule():
    try:
        clean(_name, eid=_eid)
    except Exception:
        ...


def setUp(name, eid=None):
    global _eid, _name
    _eid = eid
    _name = name

    if not _eid:
        # by default, we use local entity
        _eid = bmy.new("**/local", eid=f"test_{name}")

    bmy.build(eid=_eid)


def init(name, *, eid):
    global _local_prev_cwd, test_path

    @bmy.bmy_func()
    def init_entity(name, *, eid):
        bmy.build(eid=eid)

        with bmy.select(eid):
            path = bmy.joinpath(test_path, name, eid)
            bmy.mkdir(path)
            prev, _ = bmy.chdir(path)
            bmy.setinfo("myrrh_test", prev_path=prev)

    local_path = os.path.join(test_path, name, "local")
    os.makedirs(local_path, exist_ok=True)
    _local_prev_cwd = os.getcwd()
    os.chdir(local_path)

    if not eid:
        # by default, we use local entity
        eid = bmy.new("**/local", eid=f"test_{name}")

    init_entity(name, eid=eid)


def clean(name, *, eid):
    @bmy.bmy_func()
    def clean_entity(*, eid):
        os.chdir(_local_prev_cwd)

        with bmy.select(eid):
            prev_path = bmy.info("myrrh_test.prev_path")
            current, _ = bmy.chdir(prev_path)
            bmy.rmdir(current, force=True)

    clean_entity(eid=eid)

    local_path = os.path.join(test_path, "local", name)
    os.remove(local_path)



from myrrh.framework.mpython.mtest.msupport import _interface as Support
from myrrh.framework.mpython.mtest.msupport.mimport_helper import (
    _interface as Import_helper,
)
from myrrh.framework.mpython.mtest.msupport.mos_helper import _interface as Os_helper
from myrrh.framework.mpython.mtest.msupport.mscript_helper import (
    _interface as Script_helper,
)

support: Support
import_helper: Import_helper
os_helper: Os_helper
script_helper: Script_helper



# flake8: noqa:

from ._bmy import *
from ._bmy import __all__ as __bmy_all__

from ._bmy_exceptions import *
from ._bmy_exceptions import __all__ as __bmy_exceptions_all__

from ._bmy_internal import *
from ._bmy_internal import __all__ as __bmy_internal_all__

__all_global__ = ["debug", "init", "list_providers", "bmy_func"]
__all_manage__ = [
    "current",
    "next",
    "previous",
    "eids",
    "entity",
    "info",
    "setinfo",
    "select",
    "unselect",
    "groupkeys",
    "groupvalues",
]
__all_loading__ = ["new", "load", "build", "isbuilt", "isgroup", "ensuregroup"]
__all_state__ = ["boot", "halt", "reboot"]
__all_fs__ = [
    "chdir",
    "cp",
    "cptree",
    "edit",
    "fstat",
    "get",
    "lsdir",
    "mkdir",
    "move",
    "push",
    "pwd",
    "rm",
    "rmdir",
    "transfer",
    "joinpath",
    "realpath",
    "basename",
    "dirname",
    "abspath",
    "read",
    "write",
]
__all_exec__ = [
    "execute",
    "execute_mem_c",
    "execute_mem_ce",
    "launch",
    "kill",
    "system",
    "which",
]
__all_snaps__ = ["csnap", "desnap", "resnap", "snap", "snaps"]
__all_except__ = [
    "BmyException",
    "BmyExecutionFailure",
    "BmyInvalidParameter",
    "BmyInvalidEid",
    "BmyMyrrhFailure",
    "BmyNotReady",
    "BmyTimeout",
]

__all__ = __bmy_all__ + __bmy_exceptions_all__ + __bmy_internal_all__



__mlib__ = "AbcBuiltins"

import typing
import functools

from myrrh.core.services.system import AbcRuntime

from . import mimportlib


class AbcBuiltins(AbcRuntime):
    import builtins as local_builtins

    __frameworkpath__ = "mpython.mbuiltins"

    __all__ = [
        "ArithmeticError",
        "AssertionError",
        "AttributeError",
        "BaseException",
        "BaseExceptionGroup",
        "BlockingIOError",
        "BrokenPipeError",
        "BufferError",
        "BytesWarning",
        "ChildProcessError",
        "ConnectionAbortedError",
        "ConnectionError",
        "ConnectionRefusedError",
        "ConnectionResetError",
        "DeprecationWarning",
        "EOFError",
        "Ellipsis",
        "EncodingWarning",
        "EnvironmentError",
        "Exception",
        "ExceptionGroup",
        "False",
        "FileExistsError",
        "FileNotFoundError",
        "FloatingPointError",
        "FutureWarning",
        "GeneratorExit",
        "IOError",
        "ImportError",
        "ImportWarning",
        "IndentationError",
        "IndexError",
        "InterruptedError",
        "IsADirectoryError",
        "KeyError",
        "KeyboardInterrupt",
        "LookupError",
        "MemoryError",
        "ModuleNotFoundError",
        "NameError",
        "None",
        "NotADirectoryError",
        "NotImplemented",
        "NotImplementedError",
        "OSError",
        "OverflowError",
        "PendingDeprecationWarning",
        "PermissionError",
        "ProcessLookupError",
        "RecursionError",
        "ReferenceError",
        "ResourceWarning",
        "RuntimeError",
        "RuntimeWarning",
        "StopAsyncIteration",
        "StopIteration",
        "SyntaxError",
        "SyntaxWarning",
        "SystemError",
        "SystemExit",
        "TabError",
        "TimeoutError",
        "True",
        "TypeError",
        "UnboundLocalError",
        "UnicodeDecodeError",
        "UnicodeEncodeError",
        "UnicodeError",
        "UnicodeTranslateError",
        "UnicodeWarning",
        "UserWarning",
        "ValueError",
        "Warning",
        "WindowsError",
        "ZeroDivisionError",
        "__import__",
        "abs",
        "aiter",
        "all",
        "anext",
        "any",
        "ascii",
        "bin",
        "bool",
        "breakpoint",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "exec",
        "exit",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "quit",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "vars",
        "zip",
    ]

    __doc__ = local_builtins.__doc__

    from builtins import (  # type: ignore[misc]
        ArithmeticError,
        AssertionError,
        AttributeError,
        BaseException,
        BaseExceptionGroup,
        BlockingIOError,
        BrokenPipeError,
        BufferError,
        BytesWarning,
        ChildProcessError,
        ConnectionAbortedError,
        ConnectionError,
        ConnectionRefusedError,
        ConnectionResetError,
        DeprecationWarning,
        EOFError,
        Ellipsis,
        EncodingWarning,
        EnvironmentError,
        Exception,
        ExceptionGroup,
        FileExistsError,
        FileNotFoundError,
        FloatingPointError,
        FutureWarning,
        GeneratorExit,
        IOError,
        ImportError,
        ImportWarning,
        IndentationError,
        IndexError,
        InterruptedError,
        IsADirectoryError,
        KeyError,
        KeyboardInterrupt,
        LookupError,
        MemoryError,
        ModuleNotFoundError,
        NameError,
        NotADirectoryError,
        NotImplemented,
        NotImplementedError,
        OSError,
        OverflowError,
        PendingDeprecationWarning,
        PermissionError,
        ProcessLookupError,
        RecursionError,
        ReferenceError,
        ResourceWarning,
        RuntimeError,
        RuntimeWarning,
        StopAsyncIteration,
        StopIteration,
        SyntaxError,
        SyntaxWarning,
        SystemError,
        SystemExit,
        TabError,
        TimeoutError,
        TypeError,
        UnboundLocalError,
        UnicodeDecodeError,
        UnicodeEncodeError,
        UnicodeError,
        UnicodeTranslateError,
        UnicodeWarning,
        UserWarning,
        ValueError,
        Warning,
        ZeroDivisionError,
        abs,
        aiter,
        all,
        anext,
        any,
        ascii,
        bin,
        bool,
        breakpoint,
        bytearray,
        bytes,
        callable,
        chr,
        compile,
        complex,
        copyright,
        credits,
        delattr,
        dict,
        dir,
        divmod,
        enumerate,
        eval,
        exec,
        exit,
        filter,
        float,
        format,
        frozenset,
        getattr,
        globals,
        hasattr,
        hash,
        help,
        hex,
        id,
        input,
        int,
        isinstance,
        issubclass,
        iter,
        len,
        license,
        list,
        locals,
        map,
        max,
        memoryview,
        min,
        next,
        object,
        oct,
        ord,
        pow,
        print,
        quit,
        range,
        repr,
        reversed,
        round,
        set,
        setattr,
        slice,
        sorted,
        str,
        sum,
        super,
        tuple,
        type,
        vars,
        zip,
    )

    __io = mimportlib.module_property("io")

    importlib = mimportlib.module_property(mimportlib)

    def __import__(self, name, globals=None, locals=None, fromlist=(), level=0):
        return self.importlib.__import__(name, globals, locals, fromlist, level)

    def open(
        self,
        file,
        mode="r",
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None,
        closefd=True,
        opener=None,
    ):
        return self.__io.open(file, mode, buffering, encoding, errors, newline, closefd, opener)

    __def_builtins__: dict[str, typing.Any] = dict(__builtins__)  # type: ignore[arg-type]

    @functools.cached_property
    def __builtins__(self):
        __builtins__ = dict(self.__def_builtins__)
        __builtins__["open"] = self.open
        __builtins__["__import__"] = self.__import__

        return __builtins__


def wrap_module(module, runtimeModule, wrapped={}):
    import importlib.util

    builtins = AbcBuiltins(runtimeModule)

    def __import__(name, globals=None, locals=None, fromlist=(), level=0):
        mod = wrapped.get(name)
        if mod is None:
            mod = builtins.__import__(name, globals, locals, fromlist, level)
        return mod

    __builtins__ = dict(module.__builtins__)
    __builtins__.update(builtins.__builtins__)
    __builtins__["__import__"] = __import__

    mod = importlib.util.module_from_spec(module.__spec__)

    assert mod is not module

    mod.__builtins__ = __builtins__
    mod.__loader__.exec_module(mod)

    return mod



import sys
import threading
import fnmatch
import typing

from mplugins.provider.local import provider

from ..interfaces import (
    ICoreService,
    NoneDelegation,
    ISystem,
    ABCDelegation,
    ABCDelegationMeta,
    IRegistry,
    ICoreSnapService,
    ICoreInstanceService,
    ICoreStateService,
    ICoreShellService,
    ICoreFileSystemService,
    ICoreStreamService,
)
from ..services import cfg_init

from ...warehouse.item import NoneItem
from ...warehouse.registry import ItemRegistry
from ...provider import ServiceGroup, IProvider, Protocol, service_fullname


from ._services import FileSystemService, ShellService, StreamService
from ._registry import Registry

__all__ = [
    "Entity",
    "CoreServiceClass",
    "CoreProviderClass",
    "CoreProvider",
    "NoneShell",
    "NoneFs",
    "NoneStream",
    "NoneState",
    "NoneSnap",
    "NoneInstance",
]

_VALIDATE = cfg_init("validate_service_args", True, section="myrrh.core")

NoneShell = NoneDelegation("NoneShellType", ICoreShellService)
NoneFs = NoneDelegation("NoneFsType", ICoreFileSystemService)
NoneStream = NoneDelegation("NoneFsType", ICoreStreamService)
NoneState = NoneDelegation("NoneStateType", ICoreStateService)
NoneSnap = NoneDelegation("NoneSnapType", ICoreSnapService)
NoneInstance = NoneDelegation("NoneInstance", ICoreInstanceService)


class _ServiceGroup:
    __slots__ = ("cfg", "lock", "services", "protocols")

    cfg: IRegistry
    lock: threading.RLock
    services: dict[str, typing.Type[ICoreService]]

    def __init__(
        self,
        category: ServiceGroup,
        cfg,
        services: list[typing.Type[ICoreService]] = list(),
    ):
        self.services = {s.eref(): s for s in ServiceGroup.group(category, services)}

        self.cfg = cfg
        self.lock = threading.RLock()
        self.protocols = set()

        for serv in self.services.values():
            if not hasattr(self, str(serv.protocol)):
                setattr(self, str(serv.protocol), type("protocol", (), {}))

            protocol = getattr(self, str(serv.protocol))

            setattr(protocol, serv.name, serv())

            self.protocols.add(str(serv.protocol))


class System(_ServiceGroup, ISystem):
    shell: ShellService = NoneShell
    fs: FileSystemService = NoneFs
    stream: StreamService = NoneStream

    def __init__(self, cfg, services=list()):
        super().__init__(ServiceGroup.system, cfg, services)

        proto = getattr(self, Protocol.MYRRH.value, None)

        if proto:
            self.shell = getattr(proto, "shell", NoneShell)
            self.fs = getattr(proto, "fs", NoneFs)
            self.stream = getattr(proto, "stream", NoneStream)

        if _VALIDATE:
            self.shell = ShellService(self.shell)
            self.fs = FileSystemService(self.fs)
            self.stream = StreamService(self.stream)

    def __str__(self):
        return f'{self.cfg.eid}(System: {",".join(self.protocols)}'

    def Stream(self, protocol: str | Protocol | None = None) -> ICoreStreamService:
        if not protocol:
            return self.stream

        protocol_name = protocol.value if isinstance(protocol, Protocol) else protocol
        try:
            return getattr(self, protocol_name).stream
        except AttributeError:
            pass

        raise RuntimeError(f'Unsupported stream protocol "{protocol_name}" for {str(self)}')

    def Fs(self, protocol: str | Protocol | None = None) -> ICoreFileSystemService:
        if not protocol:
            return self.fs

        protocol_name = protocol.value if isinstance(protocol, Protocol) else protocol
        try:
            return getattr(self, protocol_name).fs
        except AttributeError:
            pass

        raise RuntimeError(f'Unsupported file system protocol "{protocol_name}" for {str(self)}')

    def Shell(self, protocol: str | Protocol | None = None) -> ICoreShellService:
        if not protocol:
            return self.shell

        protocol_name = protocol.value if isinstance(protocol, Protocol) else protocol

        try:
            return getattr(self, protocol_name).shell
        except AttributeError:
            pass

        raise RuntimeError(f'Unsupported shell protocol "{protocol_name}" for {str(self)}')


class Host(_ServiceGroup):
    state: ICoreStateService = NoneState
    snap: ICoreSnapService = NoneSnap
    inst: ICoreInstanceService = NoneInstance

    def __init__(self, cfg, services=list()):
        super().__init__(ServiceGroup.host, cfg, services)

        proto = getattr(self, Protocol.MYRRH.value, None)
        if proto:
            self.state = getattr(proto, "state", NoneState)
            self.snap = getattr(proto, "snap", NoneSnap)
            self.inst = getattr(proto, "inst", NoneInstance)


class Vendor(_ServiceGroup):
    def __init__(self, cfg, services=list()):
        super().__init__(ServiceGroup.vendor, cfg, services)


def CoreProviderClass(provider_cls, name) -> typing.Type[IProvider]:
    class _(IProvider, ABCDelegation):
        __delegated__ = (IProvider,)
        _name_ = name
        _provider_ = provider_cls

        def __init__(self, *a, **kwa):
            self.__delegate__(IProvider, self._provider_(*a, **kwa))

    return ABCDelegationMeta(_._provider_.__name__, _.__bases__, dict(_.__dict__))  # type: ignore[return-value]


def CoreServiceClass(path, serv_cls, pname):
    ServInterface = ServiceGroup[serv_cls.category].__interfaces__[serv_cls.name]

    class _(ICoreService, ServInterface, ABCDelegation):
        _serv_ = serv_cls

        _eref = "/".join((path, pname))
        category = serv_cls.category
        protocol = serv_cls.protocol
        name = serv_cls.name

        __delegated__ = (ServInterface,)

        def __init__(self, *a, **kwa):
            self.__delegate__(ServInterface, self._serv_(*a, **kwa))

        @classmethod
        def eref(cls):
            return cls._eref

    return ABCDelegationMeta(_._serv_.__name__, _.__bases__, dict(_.__dict__))


class CoreProvider(IProvider):
    SERV = 0
    CTLG = 1

    def __init__(
        self,
        providers: typing.Iterable[IProvider],
        *,
        patterns: list[str] | None = None,
    ):
        self.providers = list(providers)

        self._services: dict[str, typing.Type[ICoreService]] = dict()
        self._catalog: dict[str, IProvider] = dict()
        self.paths = ""

        self._init_using_path(providers, patterns)

    def _get_all_paths(self, providers: list[IProvider]):
        _allpaths = dict()

        for p in providers:
            serv = map(
                lambda s: CoreServiceClass(f"{self.SERV}/{service_fullname(s)}", s, p._name_),
                p.services(),
            )

            for serv in ServiceGroup.group(None, list(serv)):
                _allpaths[serv.eref()] = serv

            for item in p.catalog():
                path = f"{self.CTLG}/catalog/{item}/{p._name_}"
                _allpaths[path] = p

        return _allpaths

    def _init_using_path(self, providers, patterns):
        _allpaths = self._get_all_paths(providers)

        paths = list()
        if patterns:
            for pattern in patterns:
                paths.extend(fnmatch.filter(_allpaths, pattern))
        else:
            paths = _allpaths

        for entry in paths:
            _i, _, path, _ = entry.split("/", 3)

            match int(_i):
                case self.CTLG:
                    self._catalog[path] = _allpaths[entry]
                case self.SERV:
                    self._services[entry] = _allpaths[entry]

        def _paths_todict(path: str, dct: dict, value):
            if path:
                p, _, v = path.partition("/")
                dct[p] = _paths_todict(v, (dct.get(p) or dict()), value)
                dct[f"{p}_"] = dct.get(f"{p}_") or list()
                dct[f"{p}_"].append(v)
                return dct

            return value

        self.paths = dict()

        for path in paths:
            _paths_todict(path, self.paths, _allpaths[path])

    def services(self) -> tuple[typing.Type[ICoreService]]:
        return tuple(self._services.values())  # type: ignore[return-value]

    def catalog(self) -> tuple[str]:
        return tuple(self._catalog.keys())  # type: ignore[return-value]

    def deliver(self, name):
        provider = self._catalog.get(name, None)

        if provider:
            return provider.deliver(name)

        return NoneItem


class Entity:
    system: System
    host: Host
    vendor: Vendor

    def __new__(cls, provider: CoreProvider, items: list[ItemRegistry().WarehouseItemT] = list()):  # type: ignore[valid-type]
        self = super().__new__(cls)

        sys.__msys__["@entities@"].append(self)  # type: ignore[attr-defined]

        return self

    def __init__(self, provider: IProvider, items: list[ItemRegistry().WarehouseItemT] = list()):  # type: ignore[valid-type]
        services = provider.services()

        self.provider = provider

        self.cfg = Registry(provider, items)
        self.system = System(self.cfg, services)
        self.host = Host(self.cfg, services)
        self.vendor = Vendor(self.cfg, services)

    @property
    def eid(self):
        return self.cfg.eid

    @property
    def uuid(self):
        return self.cfg.uuid

    def __str__(self):
        return str(self.eid)

    def __repr__(self):
        return "Entity<%s>" % self.eid


setattr(Entity, ".", Entity(CoreProvider((CoreProviderClass(provider, ".")(),))))



import stat
import typing

from concurrent.futures import CancelledError

from myrrh.core.services import PID
from myrrh.core.services.system import AbcRuntime, Protocol
from myrrh.utils import mshlex


__mlib__ = "AbcWinAPI"


# copy of STARTUPINFO from python subprocess module
#
# Copyright (c) 2003-2005 by Peter Astrand <astrand@lysator.liu.se>
#
# Licensed to PSF under a Contributor Agreement.
class STARTUPINFO:
    def __init__(
        self,
        *,
        dwFlags=0,
        hStdInput=None,
        hStdOutput=None,
        hStdError=None,
        wShowWindow=0,
        lpAttributeList=None,
    ):
        self.dwFlags = dwFlags
        self.hStdInput = hStdInput
        self.hStdOutput = hStdOutput
        self.hStdError = hStdError
        self.wShowWindow = wShowWindow
        self.lpAttributeList = lpAttributeList or {"handle_list": []}

    def copy(self):
        attr_list = self.lpAttributeList.copy()
        if "handle_list" in attr_list:
            attr_list["handle_list"] = list(attr_list["handle_list"])

        return STARTUPINFO(
            dwFlags=self.dwFlags,
            hStdInput=self.hStdInput,
            hStdOutput=self.hStdOutput,
            hStdError=self.hStdError,
            wShowWindow=self.wShowWindow,
            lpAttributeList=attr_list,
        )


class AbcWinAPI(AbcRuntime):
    __frameworkpath__ = "mpython._mwinapi"

    ABOVE_NORMAL_PRIORITY_CLASS = 32768
    BELOW_NORMAL_PRIORITY_CLASS = 16384

    CREATE_BREAKAWAY_FROM_JOB = 16777216
    CREATE_DEFAULT_ERROR_MODE = 67108864
    CREATE_NO_WINDOW = 134217728
    CREATE_NEW_CONSOLE = 16
    CREATE_NEW_PROCESS_GROUP = 512
    DETACHED_PROCESS = 8
    DUPLICATE_CLOSE_SOURCE = 1
    DUPLICATE_SAME_ACCESS = 2
    ERROR_ALREADY_EXISTS = 183
    ERROR_BROKEN_PIPE = 109
    ERROR_IO_PENDING = 997
    ERROR_MORE_DATA = 234
    ERROR_NETNAME_DELETED = 64
    ERROR_NO_DATA = 232
    ERROR_NO_SYSTEM_RESOURCES = 1450
    ERROR_OPERATION_ABORTED = 995
    ERROR_PIPE_BUSY = 231
    ERROR_PIPE_CONNECTED = 535
    ERROR_SEM_TIMEOUT = 121

    FILE_FLAG_FIRST_PIPE_INSTANCE = 524288
    FILE_FLAG_OVERLAPPED = 1073741824
    FILE_GENERIC_READ = 1179785
    FILE_GENERIC_WRITE = 1179926
    FILE_MAP_ALL_ACCESS = 983071
    FILE_MAP_COPY = 1
    FILE_MAP_EXECUTE = 32
    FILE_MAP_READ = 4
    FILE_MAP_WRITE = 2
    FILE_TYPE_CHAR = 2
    FILE_TYPE_DISK = 1
    FILE_TYPE_PIPE = 3
    FILE_TYPE_REMOTE = 32768
    FILE_TYPE_UNKNOWN = 0

    GENERIC_READ = 2147483648
    GENERIC_WRITE = 1073741824
    HIGH_PRIORITY_CLASS = 128
    INFINITE = 4294967295
    INVALID_HANDLE_VALUE = 18446744073709551615
    IDLE_PRIORITY_CLASS = 64
    NORMAL_PRIORITY_CLASS = 32
    REALTIME_PRIORITY_CLASS = 256
    NMPWAIT_WAIT_FOREVER = 4294967295
    MEM_COMMIT = 4096
    MEM_FREE = 65536
    MEM_IMAGE = 16777216
    MEM_MAPPED = 262144
    MEM_PRIVATE = 131072
    MEM_RESERVE = 8192

    NULL = 0
    OPEN_EXISTING = 3

    PIPE_ACCESS_DUPLEX = 3
    PIPE_ACCESS_INBOUND = 1
    PIPE_READMODE_MESSAGE = 2
    PIPE_TYPE_MESSAGE = 4
    PIPE_UNLIMITED_INSTANCES = 255
    PIPE_WAIT = 0

    PAGE_EXECUTE = 16
    PAGE_EXECUTE_READ = 32
    PAGE_EXECUTE_READWRITE = 64
    PAGE_EXECUTE_WRITECOPY = 128
    PAGE_GUARD = 256
    PAGE_NOACCESS = 1
    PAGE_NOCACHE = 512
    PAGE_READONLY = 2
    PAGE_READWRITE = 4
    PAGE_WRITECOMBINE = 1024
    PAGE_WRITECOPY = 8

    PROCESS_ALL_ACCESS = 2097151
    PROCESS_DUP_HANDLE = 64

    SEC_COMMIT = 134217728
    SEC_IMAGE = 16777216
    SEC_LARGE_PAGES = 2147483648
    SEC_NOCACHE = 268435456
    SEC_RESERVE = 67108864
    SEC_WRITECOMBINE = 1073741824
    STARTF_USESHOWWINDOW = 1
    STARTF_USESTDHANDLES = 256
    STD_ERROR_HANDLE = 4294967284
    STD_INPUT_HANDLE = 4294967286
    STD_OUTPUT_HANDLE = 4294967285
    STILL_ACTIVE = 259
    SW_HIDE = 0
    SYNCHRONIZE = 1048576
    WAIT_ABANDONED_0 = 128
    WAIT_OBJECT_0 = 0
    WAIT_TIMEOUT = 258

    class Overlapped:
        event: int

        def GetOverlappedResult(self, __wait: bool) -> tuple[int, int]:
            raise NotImplementedError

        def cancel(self) -> None:
            raise NotImplementedError

        def getbuffer(self) -> bytes | None:
            raise NotImplementedError

    def CloseHandle(self, handle: int) -> None:
        if handle != 0:
            self.myrrh_syscall.gethandle(handle).close()

    def CreateFile(
        self,
        __file_name: str,
        __desired_access: int,
        __share_mode: int,
        __security_attributes: int,
        __creation_disposition: int,
        __flags_and_attributes: int,
        __template_file: int,
    ) -> int:
        extras = {
            "desired_access": __desired_access,
            "share_mode": __share_mode,
            "security_attributes": __security_attributes,
            "creation_disposition": __creation_disposition,
            "flags_and_attributes": __flags_and_attributes,
            "template_file": __template_file,
        }

        wiring = self.myrrh_syscall.Wiring.OFF
        if __desired_access & self.GENERIC_READ:
            wiring |= self.myrrh_syscall.Wiring.IN
        if __desired_access & self.GENERIC_WRITE:
            wiring |= self.myrrh_syscall.Wiring.OUT

        return self.myrrh_syscall.open_file(
            self.myrrh_os.fsencode(__file_name),
            wiring,
            extras=extras,
            protocol=Protocol.WINAPI,
        )

    def ConnectNamedPipe(self, handle: int, overlapped=True):
        raise NotImplementedError

    def CreateNamedPipe(
        self,
        __name: str,
        __open_mode: int,
        __pipe_mode: int,
        __max_instances: int,
        __out_buffer_size: int,
        __in_buffer_size: int,
        __default_timeout: int,
        __security_attributes: int,
    ) -> int:
        raise NotImplementedError

    def CreatePipe(self, __pipe_attrs: typing.Any, __size: int) -> tuple[int, int]:
        rd, wr = self.myrrh_syscall.open_pipe()

        return rd, wr

    def CreateJunction(self, __src_path: str, __dst_path: str) -> None:
        raise NotImplementedError

    def CreateProcess(
        self,
        __application_name: str | None,
        __command_line: str | None,
        __proc_attrs: typing.Any,
        __thread_attrs: typing.Any,
        __inherit_handles: bool,
        __creation_flags: int,
        __env_mapping: dict[
            str,
            str,
        ],
        __current_directory: str | None,
        __startup_info: STARTUPINFO,
    ) -> tuple[int, int, int, int]:
        args = mshlex.split(__command_line, False)
        if __application_name is None:
            __application_name = args[0]

        path = self.myrrh_os.p(__application_name)
        args = [self.myrrh_os.fsencode(arg) for arg in args]
        working_dir = self.myrrh_os.p(__current_directory) if __current_directory else None
        env = {self.myrrh_os.fsencode(k): self.myrrh_os.fsencode(v) for k, v in __env_mapping.items()} if __env_mapping else None

        stdin = __startup_info.hStdInput
        stdout = __startup_info.hStdOutput
        stderr = __startup_info.hStdError

        hproc = self.myrrh_syscall.open_process(
            path,
            args,
            working_dir,
            env,
            stdin,
            stdout,
            stderr,
            extras={
                "creation_flags": __creation_flags,
                "inherit_handles": __inherit_handles,
                "startup_info.flags": __startup_info.dwFlags,
                "startup_info.show_windows": __startup_info.wShowWindow,
            },
            protocol=Protocol.WINAPI,
        )

        handle = self.myrrh_syscall.gethandle(hproc, True)

        pid = handle.pid

        return hproc, 0, pid, 0

    def DuplicateHandle(
        self,
        __source_process_handle: int,
        __source_handle: int,
        __target_process_handle: int,
        __desired_access: int,
        __inherit_handle: bool,
        __options: int = 0,
    ) -> int:
        return self.myrrh_syscall.dup(__source_handle)

    def ExitProcess(self, __ExitCode: int) -> None:
        raise NotImplementedError

    def GetACP(self) -> int:
        raise NotImplementedError

    def GetFileType(self, handle: int) -> int:
        handle = self.myrrh_syscall.gethandle(handle, True)
        mode = handle.stat().st_mode
        if stat.S_ISFIFO(mode):
            return self.FILE_TYPE_CHAR
        if stat.S_ISREG(mode) or stat.S_ISDIR(mode) or stat.S_ISLNK(mode):
            return self.FILE_TYPE_DISK
        if stat.S_ISCHR(mode):
            return self.FILE_TYPE_CHAR

        return self.FILE_TYPE_UNKNOWN

    def GetCurrentProcess(self) -> int:
        return PID

    def GetExitCodeProcess(self, __process: int) -> int:
        handle = self.myrrh_syscall.gethandle(__process, True)
        return handle.exit_status

    def GetLastError(self) -> int:
        raise NotImplementedError

    def GetModuleFileName(self, __module_handle: int) -> str:
        raise NotImplementedError

    def GetStdHandle(self, __std_handle: int) -> int:
        match __std_handle:
            case self.STD_INPUT_HANDLE:
                return self.myrrh_syscall.hin
            case self.STD_OUTPUT_HANDLE:
                return self.myrrh_syscall.hout
            case self.STD_ERROR_HANDLE:
                return self.myrrh_syscall.herr

        raise ValueError(f"Invalid std handle value: {__std_handle}")

    def GetVersion(self) -> int:
        raise NotImplementedError

    def OpenProcess(self, __desired_access: int, __inherit_handle: bool, __process_id: int) -> int:
        raise NotImplementedError

    def PeekNamedPipe(self, __handle: int, __size: int = ...) -> tuple[int, int] | tuple[bytes, int, int]:
        raise NotImplementedError

    def SetNamedPipeHandleState(
        self,
        __named_pipe: int,
        __mode: int | None,
        __max_collection_count: int | None,
        __collect_data_timeout: int | None,
    ) -> None:
        raise NotImplementedError

    def TerminateProcess(self, __handle: int, __exit_code: int) -> None:
        handle = self.myrrh_syscall.gethandle(__handle, True)
        handle.terminate(extras={"exit_code": __exit_code})

    def WaitForMultipleObjects(
        self,
        __handle_seq: typing.Sequence[int],
        __wait_flag: bool,
        __milliseconds: int = ...,
    ) -> int:
        raise NotImplementedError

    def WaitForSingleObject(self, __handle: int, __milliseconds: int) -> int:
        try:
            timeout = None if __milliseconds == self.INFINITE else __milliseconds / 1000
            self.myrrh_syscall.wait(__handle, timeout)

            return self.WAIT_OBJECT_0
        except TimeoutError:
            return self.WAIT_TIMEOUT
        except CancelledError:
            return self.WAIT_ABANDONED_0
        except Exception:
            raise

    def WaitNamedPipe(self, __name: str, __timeout: int) -> None:
        raise NotImplementedError

    def ReadFile(self, handle: int, size: int, overlapped=False) -> tuple[bytes, int]:
        return (
            self.myrrh_syscall.gethandle(handle).read(size, extras={"overlapped": overlapped}),
            0,
        )

    def WriteFile(self, handle: int, buffer: bytes, overlapped=False) -> tuple[int, int]:
        return (
            self.myrrh_syscall.gethandle(handle).write(buffer, extras={"overlapped": overlapped}),
            0,
        )



import re
import functools

from myrrh.core.services.system import AbcMyrrhOs, ExecutionFailureCauseRVal
from myrrh.utils import merrno
from myrrh.utils.mstring import str2int
from myrrh.utils import mshlex

_error_table = {
    b"The system cannot find the path specified.": merrno.ENOENT,
    b"File Not Found": merrno.ENOENT,
}


class MyrrhOs(AbcMyrrhOs):
    _curdirb_ = b"."
    _pardirb_ = b".."
    _extsepb_ = b"."
    _sepb_ = b"\\"
    _pathsepb_ = b";"
    _altsepb_ = b"/"
    _defpathb_ = b".;C:\\bin"
    _devnullb_ = b"nul"
    _linesepb_ = b"\r\n"

    import ntpath

    isabs = staticmethod(ntpath.isabs)  # type: ignore[assignment]
    normpath = staticmethod(ntpath.normpath)  # type: ignore[assignment]
    joinpath = staticmethod(ntpath.join)  # type: ignore[assignment]
    basename = staticmethod(ntpath.basename)  # type: ignore[assignment]
    dirname = staticmethod(ntpath.dirname)  # type: ignore[assignment]

    __fsencoding = "utf-8"
    __encoding = "utf-8"

    def _getbinb_(self):
        return {
            b"wmic": rb"C:\Windows\System32\wbem\WMIC.exe",
            b"chcp": rb"C:\Windows\System32\chcp.com",
            b"chcp_utf8": rb"C:\Windows\System32\chcp.com 65001 > NUL &&",
            b"set": b"set",
            b"echo": b"echo",
            b"where": rb"C:\Windows\System32\where.exe",
            b"setx": rb"C:\Windows\System32\setx.exe",
            b"reg": rb"C:\Windows\System32\reg.exe",
            b"reg_utf8": rb"C:\Windows\System32\chcp.com 65001 > NUL && C:\Windows\System32\reg.exe",
            b"copy": b"copy",
            b"powershell": rb"C:\Windows\System32\WindowsPowerShell\v1.0\powershell.exe",
            b"mklink": b"mklink",
            b"cmd": rb"C:\windows\system32\cmd.exe",
            b"ping": rb"C:\Windows\System32\ping.exe",
            b"move": b"move",
            b"dir": b"dir",
            b"cd": b"cd",
            b"rename": b"rename",
            b"rmdir": b"rmdir",
            b"robocopy": rb"C:\Windows\System32\robocopy.exe",
            b"xcopy": rb"C:\Windows\System32\xcopy.exe",
            b"tar": rb"C:\Windows\System32\tar.exe",
            b"mkdir": b"mkdir",
            b"cscript": rb"C:\Windows\System32\cscript.exe",
        }

    @functools.cached_property
    def _wininfo(self):
        out, err, rval = self.cmd(b"%(wmic)s os get /all /value")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        return {a[0]: str2int(a[1]) for a in (line.split("=") for line in out.split("\r\n")) if len(a) == 2}

    def _getdefaultencoding_(self):
        cp = 0
        try:
            out, err, rval = self.cmdb(b"%(wmic)s os get CodeSet /value")
            ExecutionFailureCauseRVal(self, err, rval, 0).check()
            cp = re.search(rb"\d+", out).group().decode("utf8")  # default to utf8
            cp = str2int(cp)
        except Exception:
            pass

        return "cp%s" % cp if cp != 0 else self.__encoding

    def _localecode_(self):
        import locale

        return locale.windows_locale.get(self._wininfo["Locale"], "C")

    def _fsencoding_(self):
        return self.__fsencoding

    def _fsencodeerrors_(self):
        return "surrogatepass"

    def _getcwdb_(self):
        out, err, rval = self.cmdb(b"%(echo)s %%CD%%")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out

    def _envb_(self):
        out, err, rval = self.cmdb(b"%(set)s")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return {k.upper(): v for k, v in re.findall(b"(?P<k>[^=]*)=(?P<v>[^\\r]*)\\r\\n", out)}

    def _gettmpdirb_(self):
        out, err, rval = self.cmdb(b'%(echo)s %%TMPDIR%%&&%(echo)s %%TEMP%%&& %(echo)s %%TMP%% && (if exist "%%SystemDrive%%:\\temp" echo %%SystemDrive%%:\\temp) && (if exist "%%SystemDrive%%:\\tmp" echo %%SystemDrive%%:\\tmp)')
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        dirlist = [
            dir
            for dir in filter(
                lambda line: line not in (b"%TMPDIR%", b"%TEMP%", b"%TMP%"),
                out.splitlines(),
            )
        ]
        return dirlist[0] if dirlist else self._getcwdb_()

    def _getreadonlyenvb_(self):
        return []

    def _getdefaultshellb_(self):
        return self.getenvb().get(b"ComSpec", self._getbinb_()[b"cmd"])

    def _getdefaultshellargsb_(self):
        return (b"/c",)

    def formatshellargs(self, args, *, defaultargs=None):
        if isinstance(args, (str, bytes)):
            args = (self.shencode(args),)

        return (*defaultargs, *args)

    def getshellscriptb(self, script):
        codepage = self.defaultencoding
        codepage = codepage[2:] if codepage.startswith("cp") else "65001"  # default to utf8
        return b"%s %s 1>NUL 2>NUL & " % (self.getbinb.get(b"chcp", b"chcp"), codepage.encode()) + script

    def sh_escape_bytes(self, bytes):
        return mshlex.winshell_escape_for_cmd_exe_b(bytes)

    def environkeyformat(self, key):
        return key.upper()

    def error_translate(self, error):
        if error:
            error = self.shencode(error)
            errno = _error_table.get(error, merrno.default_errorno)
            if errno == merrno.default_errorno:
                return merrno.errno_from_msgb(error)
            return errno
        return merrno.default_errorno

    def syspathb(self, path=None):
        if path == b"nul":
            return b"\\\\.\\nul"

        return super().syspathb(path)



from myrrh.core.services.system import AbcRuntime

__mlib__ = "AbcMsvcrt"


class AbcMsvcrt(AbcRuntime):
    __frameworkpath__ = "mpython.mmsvcrt"

    def open_osfhandle(self, __handle: int, __flags: int) -> int:
        return __handle

    def get_osfhandle(self, __handle: int) -> int:
        return self.myrrh_syscall.gethandle(__handle).detach()



import os

from myrrh.core.services.system import AbcRuntime


class Ansimap(AbcRuntime):
    CSI = {
        b"H": b"\033[A",
        b"P": b"\033[B",
        b"M": b"\033[C",
        b"K": b"\033[D",
        b"I": b"\033[S",
        b"Q": b"\033[T",
        b"R": b"\033[1P",
        b"S": b"\x7F",
        b"G": b"\033[H",
        b"O": b"\033[F",
    }

    def __init__(self, iterator, abortkey=None):
        self.map = self.__wintoposix if os.name == "nt" else self.__ident
        self.__iter = iterator
        self.__abortkey = abortkey
        self.__keys = []

    def __iter__(self):
        return self

    def __next__(self):
        return self.map()

    def __call__(self):
        return self.map()

    def __ident(self):
        if len(self.__keys) == 0:
            k = next(self.__iter)
            self.__keys.extend(k)

        k = self.__keys.pop(0)

        if k == self.__abortkey:
            raise StopIteration

        return k

    def __wintoposix(self):
        if len(self.__keys) == 0:
            k = next(self.__iter)
            self.__keys.extend(k)

        k = self.__keys.pop(0)
        ordk = ord(k)
        if ordk == 224:
            k = next(self.__iter)
            self.__keys += list(Ansimap.CSI.get(k, "\0"))
            k = chr(self.__keys.pop(0)).encode()

        if ordk == 0:
            next(self.__iter)
            k = ""

        if k == "\r":
            k = "\n"
        if k == self.__abortkey:
            raise StopIteration

        return k



import functools
import threading
import warnings
import weakref
import pydantic

from ..interfaces import IRegistry
from ..services import cfg_init

from ... import provider
from ...warehouse.item import NoneItem
from ...warehouse.registry import ItemRegistry


class Registry(IRegistry):
    __eid__: dict[str, weakref.WeakSet] = dict()
    _item_updaters: dict[str, list[provider.IProvider]]

    def __init__(self, provider: provider.IProvider, items: list[ItemRegistry().WarehouseItemT]):  # type: ignore[valid-type]
        super().__init__()

        self.lock = threading.RLock()
        self._item_updaters = dict()
        self._predefined_items = {item.type_: item for item in items}

        for name in provider.catalog():
            updaters = self._item_updaters.get(name) or list()
            updaters.append(provider)
            self._item_updaters[name] = updaters
            self[name] = NoneItem

    def __setattr__(self, name, value):
        if name != "lock" and not name.startswith("_"):
            self[name] = value
            return

        super().__setattr__(name, value)

    def __getattr__(self, name):
        try:
            if not name.startswith("_"):
                return self[name]
        except KeyError:
            pass

        raise AttributeError(name)

    def __getitem__(self, name):
        with self.lock:
            name, _, next = name.partition(".")
            item = super().get(name, NoneItem)
            if not item:
                updaters = self._item_updaters.get(name) or list()
                for p in updaters:
                    try:
                        self.append(p.deliver(name))
                    except Exception as e:
                        warnings.warn(
                            'a provider failed to provide a required item "%s" : %s' % (name, e),
                            UserWarning,
                        )

                item = super().get(name, NoneItem)

            if next:
                item = item.__getitem__(next)

            return item

    def predefined(self):
        return list(self._predefined_items.values())

    def defined_values(self):
        return list(filter(None, super().values()))

    def values(self):
        return list(self[item] for item in self)

    def get(self, key, default=NoneItem):
        return self[key] or default

    def append(self, item, mode="update"):
        if not item:
            return

        if isinstance(item, dict):
            item = pydantic.TypeAdapter(ItemRegistry().WarehouseItemT).validate_python(item)

        try:
            cur_item = super().__getitem__(item.type_)
        except KeyError:
            cur_item = None

        if not cur_item:
            cur_item = self._predefined_items.get(item.type_)
            mode = "keep"

        if not cur_item and mode in ("update", "keep"):
            mode = "replace"

        if mode == "update":
            dump = cur_item.model_dump()
            dump.update(item.model_dump(exclude_unset=True))
            self[item.type_] = item.model_validate(dump)
            return

        if mode == "keep":
            dump = item.model_dump()
            dump.update(cur_item.model_dump(exclude_unset=True))
            self[item.type_] = item.model_validate(dump)

            return

        if mode == "replace":
            self[item.type_] = item
            return

        raise ValueError("invalid mode argument, must be one of 'replace', 'preserve' or 'force' , not %s" % mode)

    @functools.cached_property
    def eid(self):
        name = str(self.id.id)
        if not name:
            name = cfg_init("eid_prefix", "entity", section="myrrh.core.entity")

        # set unique entity id
        with self.lock:
            eids = self.__eid__.get(name) or list()
            eid = f"{name}/{len(eids)}" if len(eids) else name
            eids.append(weakref.ref(self))
            self.__eid__[name] = eids

        return eid

    @functools.cached_property
    def uuid(self):
        if self.id.uuid:
            return str(self.id.uuid)

        import uuid

        return str(uuid.uuid4())



import errno
import os
import subprocess
import typing
import signal
import shutil


from myrrh.utils import mshlex
from myrrh.provider import (
    Wiring,
    Protocol,
    Whence,
    Stat,
    IShellService,
    IFileSystemService,
    IStreamService,
    StatField,
)

from myrrh.utils.mhandle import LightHandler

from myrrh.core.services import cfg_init


def _working_dir(cwd):
    cwd = os.fsdecode(cwd) if isinstance(cwd, bytes) else cwd
    return None if cwd is None else os.path.expanduser("~") if len(cwd) == 0 else os.path.join(os.path.expanduser("~"), cwd) if cwd in [os.path.curdir, os.path.pardir] else cwd


_shell = None
_shellargs: tuple[str] | None = None

if os.name == "nt":
    _shell = os.environ.get("COMSPEC", "cmd.exe")
    _shell = shutil.which(_shell)
    _shellarg = "/c"

if os.name == "posix":
    _shell = shutil.which("sh")
    _shellarg = "-c"

_default_shell_encoding = "utf8"

OSErrorEBADF = OSError(errno.EBADF, os.strerror(errno.EBADF))


class Shell(IShellService):
    protocol = Protocol.MYRRH

    def execute(
        self,
        command,
        working_dir=None,
        env=None,
        *,
        extras: dict[str, typing.Any] | None = None,
    ):
        working_dir = _working_dir(working_dir)
        if isinstance(command, list):
            command = b" ".join(command)

        command = os.fsdecode(command)
        env = env if env is None else {os.fsdecode(k): os.fsdecode(v) for k, v in env.items()}

        with subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            shell=True,
            cwd=working_dir,
            env=env,
            text=False,
        ) as process:
            out, err = process.communicate()
            return out, err, process.returncode

    def spawn(
        self,
        command,
        working_dir=None,
        env=None,
        *,
        extras: dict[str, typing.Any] | None = None,
    ):
        working_dir = _working_dir(working_dir)

        command = list(map(os.fsdecode, command))

        env = env if env is None else {os.fsdecode(k): os.fsdecode(v) for k, v in env.items()}

        proc = subprocess.Popen(command, cwd=working_dir, env=env)

        if proc is None:
            raise Exception("can't execute command %s" % command)

        return proc.pid

    def signal(self, proc_id, sig=signal.SIGTERM):
        os.kill(proc_id, sig)


class FileSystem(IFileSystemService):
    """
    Service handle all file system function
    """

    protocol = Protocol.MYRRH

    def exist(self, file_path, *, extras: dict[str, typing.Any] | None = None):
        return os.path.exists(file_path)

    def rm(self, file_path, *, extras: dict[str, typing.Any] | None = None):
        os.remove(file_path)

    def mkdir(self, path, *, extras: dict[str, typing.Any] | None = None):
        if os.path.isdir(path):
            return
        os.makedirs(path)

    def rmdir(self, path, *, extras: dict[str, typing.Any] | None = None):
        os.rmdir(path)

    def is_container(self, path, *, extras: dict[str, typing.Any] | None = None):
        return os.path.isdir(path)

    def list(self, path: bytes, *, extras: dict[str, typing.Any] | None = None):
        raise NotImplementedError

    def stat(self, path: bytes, *, extras: dict | None = None) -> dict:
        stat = os.stat(path)
        return {
            "st_mode": stat.st_mode,
            "st_ino": stat.st_ino,
            "st_dev": stat.st_dev,
            "st_nlink": stat.st_nlink,
            "st_uid": stat.st_uid,
            "st_gid": stat.st_gid,
            "st_size": stat.st_size,
            "st_atime": stat.st_atime,
            "st_mtime": stat.st_mtime,
            "st_ctime": stat.st_ctime,
        }


try:
    import _winapi as winapi
    import msvcrt
except ImportError:
    winapi = False  # type: ignore[assignment]


class StreamPosix(IStreamService):
    protocol = Protocol.POSIX
    chunk_sz = cfg_init("rd_chunk_size", 2048, section="mplugins.provider.local")

    handler = LightHandler()

    def open_file(self, path: bytes, wiring: int, *, extras: dict | None = None) -> tuple[bytes, int]:
        try:
            flags = 0
            mode = 511

            wiring_ = Wiring(wiring)

            if wiring_ & Wiring.INOUT:
                flags |= os.O_RDWR
                if not wiring_ & Wiring.RESET:
                    flags |= os.O_APPEND
            elif wiring_ & Wiring.IN:
                flags |= os.O_RDONLY
            elif wiring_ & Wiring.OUT:
                flags |= os.O_WRONLY
                if not wiring_ & Wiring.RESET:
                    flags |= os.O_APPEND

            if wiring_ & Wiring.CREATE:
                flags |= os.O_CREAT

            if hasattr(os, "O_BINARY"):
                flags |= os.O_BINARY

            if extras:
                flags = extras.get("flags", wiring)
                mode = extras["mode"]

            fd = os.open(os.fsdecode(path), flags, mode)
            handle = self.handler.new(path, fd, None)

            return path, handle

        except KeyError as e:
            raise TypeError(f"open() missing required argument {e}") from None

    def open_process(
        self,
        path: bytes,
        wiring: int,
        args: list[bytes],
        working_dir: bytes | None = None,
        env: dict[bytes, bytes] | None = None,
        *,
        extras: dict | None = None,
    ) -> tuple[bytes, int, int, int, int]:
        import subprocess

        stdin = stdout = stderr = None
        process_group = gid = gids = uid = None
        creationflags = subprocess.CREATE_NEW_PROCESS_GROUP if winapi else 0  # type: ignore[attr-defined]
        umask = -1
        wiring = Wiring(wiring)

        if wiring & Wiring.IN:
            stdin = subprocess.PIPE
        if wiring & Wiring.OUT:
            stdout = subprocess.PIPE
        if wiring & Wiring.ERR:
            stderr = subprocess.PIPE

        if extras:
            process_group = extras.get("process_group")
            gid = extras.get("gid")
            gids = extras.get("gids")
            uid = extras.get("uid")
            umask = extras.get("umask") or umask
            creationflags = extras.get("creationflags") or creationflags

        env_: dict[str, str] | None = env and {os.fsencode(k): os.fsencode(v) for k, v in env.items()}  # type: ignore[assignment]

        proc = subprocess.Popen(executable=path, args=args, cwd=working_dir, env=env_, stdin=stdin, stdout=stdout, stderr=stderr, creationflags=creationflags, process_group=process_group, group=gid, extra_groups=gids, user=uid, umask=umask, close_fds=False)  # type: ignore[misc]

        if proc.stdin:
            hin = self.handler.new(path, proc.stdin.fileno(), proc.stdin)
        else:
            hin = None

        if proc.stdout:
            hout = self.handler.new(path, proc.stdout.fileno(), proc.stdout)

        else:
            hout = None

        if proc.stderr:
            herr = self.handler.new(path, proc.stderr.fileno(), proc.stderr)
        else:
            herr = None

        hproc = self.handler.new(path, -1, proc)

        return path, hproc, hin, hout, herr

    def read(self, handle: int, nbytes: int, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        fd = self.handler.h(handle, 1)

        return bytearray(os.read(fd, nbytes))

    def readall(self, handle: int, *, extras: dict | None = None) -> bytearray:
        fd = self.handler.h(handle, 1)

        sz = os.stat(fd).st_size
        return bytearray(os.read(fd, sz))

    def readchunk(self, handle: int, *, extras: dict | None = None) -> bytearray:
        fd = self.handler.h(handle, 1)

        return bytearray(os.read(fd, self.chunk_sz))

    def write(self, handle: int, data: bytes, *, extras: dict[str, typing.Any] | None = None):
        fd = self.handler.h(handle, 1)

        return os.write(fd, data)

    def close(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> None:
        try:
            manager = self.handler.h(handle, 2)
        except OSError:
            manager = None

        fd = self.handler.close(handle, 1)
        try:
            if fd > 0:
                if manager and hasattr(manager, "close"):
                    manager.close()
                else:
                    os.close(fd)
        except OSError:
            pass

    def seek(
        self,
        handle: int,
        pos: int,
        whence: int,
        *,
        extras: dict[str, typing.Any] | None = None,
    ) -> int:
        fd = self.handler.h(handle, 1)
        whence = Whence(whence)

        match whence:
            case Whence.SEEK_CUR:
                whence_ = os.SEEK_CUR
            case Whence.SEEK_SET:
                whence_ = os.SEEK_SET
            case Whence.SEEK_END:
                whence_ = os.SEEK_END

        return os.lseek(fd, pos, whence_)

    def sync(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> None:
        fd = self.handler.h(handle, 1)

        return os.fsync(fd)

    def flush(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        self.handler.h(handle, 1)

        return bytearray()

    def truncate(self, handle: int, length: int, *, extras: dict | None = None) -> None:
        fd = self.handler.h(handle, 1)
        return os.truncate(fd, length)

    def stat(
        self,
        handle: int,
        fields: int = StatField.ALL.value,
        *,
        extras: dict | None = None,
    ) -> dict:
        path = self.handler.h(handle, 0)

        st = Stat()

        exit_status = None
        pid = None

        fields = StatField(fields)

        if fields & StatField.STATUS or fields & StatField.PID:
            try:
                proc: subprocess.Popen = self.handler.h(handle, 2)

                pid = proc.pid

                try:
                    exit_status = proc._wait(0)  # type: ignore[attr-defined]
                except subprocess.TimeoutExpired:
                    pass

            except OSError:
                pass

        if fields & ~(StatField.STATUS | StatField.PID):
            st = os.stat(path)  # type: ignore[assignment]

        return Stat._make(
            (
                st.st_mode,
                st.st_ino,
                st.st_dev,
                st.st_nlink,
                st.st_uid,
                st.st_gid,
                st.st_size,
                st.st_atime,
                st.st_mtime,
                st.st_ctime,
                pid,
                exit_status,
            )
        )._asdict()

    def terminate(self, handle: int, *, extras: dict | None = None) -> None:
        proc: subprocess.Popen = self.handler.h(handle, 2)
        if winapi:
            subprocess.check_call(f'cmd /C "taskkill /pid {proc.pid} /F /T 1> nul 2> nul"')
        proc.terminate()


if winapi:

    class StreamWinAPi(IStreamService):
        protocol = Protocol.WINAPI
        chunk_sz = cfg_init("rd_chunk_size", 2048, section="mplugins.provider.local")

        handles: dict[int, typing.Any] = dict()

        def _h(self, handle) -> tuple[bytes, int]:
            info = self.handles[handle]
            if info is None:
                raise OSErrorEBADF

            return handle, *info

        def open_file(self, path: bytes, wiring: int, *, extras: dict | None = None) -> tuple[bytes, int]:
            extras = extras or dict()
            wiring = Wiring(wiring)

            desired_access = 0
            create_disposition = 3  # open existing
            file_attributes = 0x80  # FILE_ATTRIBUTE_NORMAL
            template_file = 0  # no template

            if wiring & Wiring.IN:
                desired_access |= winapi.GENERIC_READ  # type: ignore[attr-defined]
            if wiring & Wiring.OUT:
                desired_access |= winapi.GENERIC_READ  # type: ignore[attr-defined]
            if wiring & Wiring.CREATE:
                if wiring & Wiring.RESET:
                    create_disposition = 2  # create always
                else:
                    create_disposition = 1  # create new
            path = os.fsdecode(path)
            handle = winapi.CreateFile(  # type: ignore[attr-defined]
                path,
                extras.get("desired_access", desired_access),
                extras.get("share_mode", 0),
                extras.get("security_attributes", 0),
                extras.get("creation_disposition", create_disposition),
                extras.get("flags_and_attributes", file_attributes),
                extras.get("template_file", template_file),
            )

            self.handles[handle] = (path, None)

            return os.fsencode(path), handle

        def open_process(
            self,
            path: bytes,
            wiring: int,
            args: list[bytes],
            working_dir: bytes | None = None,
            env: dict[bytes, bytes] | None = None,
            *,
            extras: dict | None = None,
        ) -> tuple[bytes, int, int | None, int | None, int | None]:
            args = mshlex.list2cmdlineb(args)
            args = os.fsdecode(args)
            wiring = Wiring(wiring)

            extras = extras or {}

            working_dir = _working_dir(working_dir)
            env = env if env is None else {os.fsdecode(k): os.fsdecode(v) for k, v in env.items()}

            creation_flags = extras.get("creation_flags", 0)

            stdin_r = stdin_w = stdout_r = stdout_w = stderr_r = stderr_w = None
            stdin_proc = stdout_proc = stderr_proc = None
            toclose = list()
            tokeep = list()
            if wiring & Wiring.IN:
                stdin_r, stdin_w = winapi.CreatePipe(None, 0)  # type: ignore[attr-defined]
                toclose.append(stdin_r)
                tokeep.append(stdin_w)
                stdin_proc = winapi.DuplicateHandle(winapi.GetCurrentProcess(), stdin_r, winapi.GetCurrentProcess(), 0, 1, winapi.DUPLICATE_SAME_ACCESS)  # type: ignore
                toclose.append(stdin_proc)

            if wiring & Wiring.OUT:
                stdout_r, stdout_w = winapi.CreatePipe(None, 0)  # type: ignore[attr-defined]
                toclose.append(stdout_w)
                tokeep.append(stdout_r)
                stdout_proc = winapi.DuplicateHandle(winapi.GetCurrentProcess(), stdout_w, winapi.GetCurrentProcess(), 0, 1, winapi.DUPLICATE_SAME_ACCESS)  # type: ignore
                toclose.append(stdout_proc)

            if wiring & Wiring.ERR:
                stderr_r, stderr_w = winapi.CreatePipe(None, 0)  # type: ignore[attr-defined]
                toclose.append(stderr_w)
                tokeep.append(stderr_r)
                stderr_proc = winapi.DuplicateHandle(winapi.GetCurrentProcess(), stderr_w, winapi.GetCurrentProcess(), 0, 1, winapi.DUPLICATE_SAME_ACCESS)  # type: ignore
                toclose.append(stderr_proc)

            startup_info = subprocess.STARTUPINFO(  # type: ignore[attr-defined]
                dwFlags=extras.get("startup_info.flags", 0 if not wiring else winapi.STARTF_USESTDHANDLES), hStdInput=stdin_proc, hStdOutput=stdout_proc, hStdError=stderr_proc, lpAttributeList={"handle_list": []}, wShowWindow=extras.get("startup_info.show_window", 0)  # type: ignore[attr-defined]
            )

            env_ = env and {os.fsdecode(k): os.fsdecode(v) for k, v in env.items()} or dict()

            try:
                hp, ht, pid, _tid = winapi.CreateProcess(None, args, None, None, int(wiring.value != 0), creation_flags, env_, working_dir, startup_info)  # type: ignore

                winapi.CloseHandle(ht)  # type: ignore[attr-defined]

            except:  # noqa: E722
                for h in tokeep:
                    winapi.CloseHandle(h)  # type: ignore[attr-defined]

                raise

            finally:
                for h in toclose:
                    winapi.CloseHandle(h)  # type: ignore[attr-defined]

            self.handles[hp] = (path, pid)
            if stdin_w is not None:
                self.handles[stdin_w] = (path, pid)
            if stdout_r is not None:
                self.handles[stdout_r] = (path, pid)
            if stderr_r is not None:
                self.handles[stderr_r] = (path, pid)

            return path, hp, stdin_w, stdout_r, stderr_r

        def read(
            self,
            handle: int,
            nbytes: int,
            *,
            extras: dict[str, typing.Any] | None = None,
        ) -> bytearray:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            data, _ = winapi.ReadFile(handle, nbytes, False)  # type: ignore[attr-defined]
            return bytearray(data)

        def readall(self, handle: int, *, extras: dict | None = None) -> bytearray:
            handle, _, _ = self._h(handle)  # type: ignore[misc]

            raise NotImplementedError

        def readchunk(self, handle: int, *, extras: dict | None = None) -> bytearray:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            return self.read(handle, self.chunk_sz, extras=extras)

        def write(
            self,
            handle: int,
            data: bytes,
            *,
            extras: dict[str, typing.Any] | None = None,
        ):
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            sz, _ = winapi.WriteFile(handle, data, False)  # type: ignore[attr-defined]
            return sz

        def close(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> None:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            self.handles.pop(handle)
            winapi.CloseHandle(handle)  # type: ignore[attr-defined]

        def sync(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> None:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            fd = msvcrt.open_osfhandle(handle, os.O_RDONLY)  # type: ignore[attr-defined]

            os.fsync(fd)

        def flush(self, handle: int, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            return bytearray()

        def truncate(
            self,
            handle: int,
            length: int | None = None,
            *,
            extras: dict[str, typing.Any] | None = None,
        ) -> None:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            raise NotImplementedError

        def seek(
            self,
            handle: int,
            pos: int,
            whence: int,
            *,
            extras: dict[str, typing.Any] | None = None,
        ) -> int:
            handle, _, _ = self._h(handle)  # type: ignore[misc]
            whence = Whence(whence)

            match whence:
                case Whence.SEEK_CUR:
                    whence_ = os.SEEK_CUR
                case Whence.SEEK_SET:
                    whence_ = os.SEEK_SET
                case Whence.SEEK_END:
                    whence_ = os.SEEK_END

            fd = msvcrt.open_osfhandle(handle, os.O_RDONLY)  # type: ignore[attr-defined]

            return os.lseek(fd, pos, whence_)

        def stat(
            self,
            handle: int,
            fields: int = StatField.ALL.value,
            *,
            extras: dict | None = None,
        ) -> dict:
            st = Stat()
            exit_status = None
            fields = StatField(fields)

            handle, path, pid = self._h(handle)  # type: ignore[misc]

            if fields & ~(StatField.STATUS | StatField.PID):
                st = os.stat(path)  # type: ignore[assignment]
                exit_status = None

            if pid and fields & StatField.STATUS:
                exit_code = winapi.GetExitCodeProcess(handle)  # type: ignore[attr-defined]
                if exit_code != winapi.STILL_ACTIVE:  # type: ignore[attr-defined]
                    exit_status = exit_code

            return Stat._make(
                (
                    st.st_mode,
                    st.st_ino,
                    st.st_dev,
                    st.st_nlink,
                    st.st_uid,
                    st.st_gid,
                    st.st_size,
                    st.st_atime,
                    st.st_mtime,
                    st.st_ctime,
                    pid,
                    exit_status,
                )
            )._asdict()

        def terminate(self, handle: int, *, extras: dict | None = None) -> None:
            try:
                exit_code = extras and extras.get("exit_code", 1) or 1
                winapi.TerminateProcess(handle, exit_code)  # type: ignore[attr-defined]
            except PermissionError:
                if winapi.GetExitCodeProcess(handle) == winapi.STILL_ACTIVE:  # type: ignore[attr-defined]
                    raise

    class StreamSystemAPI(StreamWinAPi):
        ...

else:

    class StreamSystemAPI(StreamPosix):  # type: ignore[no-redef]
        ...


class Stream(StreamPosix):
    protocol = Protocol.MYRRH



from types import ModuleType

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from myrrh.framework.mpython import mbuiltins

__mlib__ = "AbcImportHelper"


class _interface(ABC):
    import test.support.import_helper as local_import_helper

    @abstractmethod
    def import_module(self, name, deprecated=False, *, required_on=()) -> ModuleType:
        ...

    @abstractmethod
    def make_legacy_pyc(self, name, deprecated=False, *, required_on=()) -> ModuleType:
        ...


class AbcImportHelper(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtest.msupport.mimport_helper"

    __delegated__ = {_interface: _interface.local_import_helper}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_import_helper, self)
        mod.make_legacy_pyc = self._unsupported

        self.__delegate__(_interface, mod)

    def _unsupported(self, *a, **kwa):
        raise NotImplementedError("not supported by myrrh")
